""" Unit Tests for Py-ART's graph/gridmapdisplay_basemap.py module. """
# execute this script to create figure_gridmapdisplay_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import matplotlib.pyplot as plt
import pytest

import pyart

RESOLUTION = 'c'    # crude resolution to speed up tests


@pytest.mark.skipif(not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
                    reason="Basemap is not installed.")
def test_gridmapdisplay_simple_basemap(outfile=None):
    # test basic GridMapDisplay functionality
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_basemap(ax=ax, resolution=RESOLUTION)
    display.plot_grid('reflectivity', vmin=-5., vmax=35)
    if outfile:
        fig.savefig(outfile)


@pytest.mark.skipif(not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
                    reason="Basemap is not installed.")
def test_gridmapdisplay_fancy_basemap(outfile=None):
    # test a bunch of GridMapDisplay functionality
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    display.debug = True
    fig = plt.figure()

    ax = fig.add_subplot(331)
    display.plot_basemap(ax=ax, resolution=RESOLUTION)
    display.plot_grid(
        'reflectivity', vmin=-5., vmax=35, mask_outside=True,
        axislabels_flag=True, axislabels=('foo', 'bar'), title='Special title')
    display.plot_crosshairs(line_style='b--')

    ax = fig.add_subplot(332)
    display.plot_grid('reflectivity', axislabels_flag=True)

    ax = fig.add_subplot(333)
    display.plot_colorbar()

    ax = fig.add_subplot(334)
    display.plot_latitude_slice('reflectivity', mask_outside=True)

    ax = fig.add_subplot(335)
    display.plot_latitude_slice('reflectivity', title='Lat title')

    ax = fig.add_subplot(336)
    grid.fields['reflectivity']['valid_min'] = 0.
    grid.fields['reflectivity']['valid_max'] = 30.
    display.plot_longitude_slice('reflectivity', mask_outside=True)

    ax = fig.add_subplot(337)
    display.plot_longitude_slice('reflectivity', title='Lon title')

    ax = fig.add_subplot(338)
    del display.grid.fields['reflectivity']['long_name']
    display.plot_colorbar()

    if outfile:
        fig.savefig(outfile)


@pytest.mark.skipif(not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
                    reason="Basemap is not installed.")
def test_plot_basemap_not_using_corners(outfile=None):
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_basemap(
        ax=ax, resolution=RESOLUTION, max_lon=None, auto_range=False)


@pytest.mark.skipif(not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
                    reason="Basemap is not installed.")
def test_generate_filename_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    filename = display.generate_filename('reflectivity', 0)
    assert isinstance(filename, str)


@pytest.mark.skipif(not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
                    reason="Basemap is not installed.")
def test_generate_titles_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)

    title = display.generate_longitudinal_level_title('reflectivity', 0)
    assert isinstance(title, str)

    title = display.generate_latitudinal_level_title('reflectivity', 0)
    assert isinstance(title, str)


@pytest.mark.skipif(not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
                    reason="Basemap is not installed.")
def test_get_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    basemap = display.get_basemap()


@pytest.mark.skipif(not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
                    reason="Basemap is not installed.")
def test_error_raising_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)

    # no mappable
    pytest.raises(ValueError, display.plot_colorbar)

    # no field
    display.mappables.append(None)  # mock the mappable
    pytest.raises(ValueError, display.plot_colorbar)


if __name__ == "__main__":
    test_gridmapdisplay_simple('figure_gridmapdisplay_simple_basemap.png')
    test_gridmapdisplay_fancy('figure_gridmapdisplay_fancy_basemap.png')



""" Unit Tests for Py-ART's _debug_info module. """

import sys
import warnings

import pyart
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def test_debug_info():
    # test to see that something is written when _debug_info is called
    # we don't care what is written, just that something is.
    buf = StringIO()
    pyart._debug_info(buf)
    assert len(buf.getvalue()) > 0


def test_debug_stdout():
    # lack of error is assumed to mean that call succeed
    pyart._debug_info()


# Class for Mocking ImportErrors from
# http://stackoverflow.com/questions/2481511/mocking-importerror-in-python
class DisableModules(object):

    def __init__(self, modules):
        self.modules = modules

    def find_module(self, fullname):
        if fullname in self.modules:
            raise ImportError('Debug import failure for %s' % fullname)


def test_debug_info_all_disabled():
    modules = ['numpy', 'scipy', 'matplotlib', 'netCDF4', 'cylp', 'glpk',
               'cvxopt', 'mpl_toolkits', 'platform']
    for module in modules:
        if module in sys.modules:
            del sys.modules[module]
    fail_loader = DisableModules(modules)
    sys.meta_path.append(fail_loader)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        # test to see that something is written when _debug_info is called
        # we don't care what is written, just that something is.
        buf = StringIO()
        pyart._debug_info(buf)
        assert len(buf.getvalue()) > 0
    # remove the Mocked ImportErrors
    sys.meta_path.remove(fail_loader)



"""
Configuration file for the Python ARM Radar Toolkit (Py-ART)

The values for a number of Py-ART parameters and the default metadata created
when reading files, correcting fields, etc. is controlled by this single
Python configuration file.

Py-ART's configuration can be modified by setting the environment variable
PYART_CONFIG to point to a configuration file with formatting similar to this
file.

The recommended method for changing these defaults is for users to copy this
file into their home directory, rename it to .pyart_config.py, make any
desired changes, and adjust their login scripts to set the PYART_CONFIG
environment variable to point to .pyart_config.py in their home directory.

Py-ART's configuration can also be modified within a script or shell session
using the load_config functions, such modification will last until the end of
the script/session or until a new configuration is loaded.

"""

##############################################################################
##############################################################################
# Simple configuration
#
# Adjust the values of the variable (right hand side of the equal sign) in
# this section for an easy method of customizing Py-ART.  Do not change the
# variable names (the left hand side of the equal sign). More advanced
# settings are based upon these variables.  Most users will find that
# adjusting this section is all that is needed.
##############################################################################
##############################################################################

# The default fill value for masked arrays and _FillValue keys
fill_value = -9999.0

# Field names used when reading in radar files and in the various correction
# and retrieval algorithms. The comments in this section provide additional
# information about the fields in that section.

# Radar reflectivity fields, DZ
reflectivity = 'reflectivity'
corrected_reflectivity = 'corrected_reflectivity'
total_power = 'total_power'

# Mean Doppler velocity fields, VEL
velocity = 'velocity'
corrected_velocity = 'corrected_velocity'
simulated_velocity = 'simulated_velocity'

# Spectral width fields, SW
spectrum_width = 'spectrum_width'

# Differential reflectivity fields, ZDR
differential_reflectivity = 'differential_reflectivity'
corrected_differential_reflectivity = 'corrected_differential_reflectivity'

# Cross correlation ratio, correlation coefficient, RhoHV
cross_correlation_ratio = 'cross_correlation_ratio'

# Normalized coherent power, signal quality index, SQI, NCP
normalized_coherent_power = 'normalized_coherent_power'

# Differential phase shift, PhiDP
differential_phase = 'differential_phase'
unfolded_differential_phase = 'unfolded_differential_phase'
corrected_differential_phase = 'corrected_differential_phase'

# Specific differential phase shift, KDP
specific_differential_phase = 'specific_differential_phase'
corrected_specific_differential_phase = 'corrected_specific_differential_phase'

# Linear depolarization ration (h - horizontal, v - vertical), LDR
linear_depolarization_ratio = 'linear_polarization_ratio'
linear_depolarization_ratio_h = 'linear_depolarization_ratio_h'
linear_depolarization_ratio_v = 'linear_depolarization_ratio_v'

# Misc fields
signal_to_noise_ratio = 'signal_to_noise_ratio'
rain_rate = 'rain_rate'
radar_estimated_rain_rate = 'radar_estimated_rain_rate'
radar_echo_classification = 'radar_echo_classification'
specific_attenuation = 'specific_attenuation'
specific_differential_attenuation = 'specific_differential_attenuation'

# visibility processing fields
terrain_altitude = 'terrain_altitude'
bent_terrain_altitude = 'bent_terrain_altitude'
terrain_slope = 'terrain_slope'
terrain_aspect = 'terrain_aspect'
elevation_angle = 'elevation_angle'
visibility = 'visibility'
min_vis_altitude = 'min_vis_altitude'
min_vis_elevation = 'min_vis_elevation'
incident_angle = 'incident_angle'
effective_area = 'effective_area'
sigma_0 = 'sigma_0'
rcs_clutter = 'rcs_clutter'
dBm_clutter = 'dBm_clutter'
dBZ_clutter = 'dBZ_clutter'
visibility_polar = 'visibility_polar'

# Textures
differential_phase_texture = 'differential_phase_texture'

# Wind retrieval fields
eastward_wind_component = 'eastward_wind_component'
northward_wind_component = 'northward_wind_component'
vertical_wind_component = 'vertical_wind_component'

# profile variables
height = 'height'
interpolated_profile = 'interpolated_profile'
height_over_iso0 = 'height_over_iso0'

# path integrated attenuation
path_integrated_attenuation = 'path_integrated_attenuation'
path_integrated_differential_attenuation = ('path_integrated' +
                                            'differential_attenuation')

# End of Simple Configuration section

##############################################################################
##############################################################################
# Advanced Configuration
#
# Most users will not want to make any changes in this section.  For users
# who want a more fine-grained control over Py-ART's configuration this
# section provides access to these controls.  The layout of this section can
# be changed, the only requirement for a valid configuration file is that
# the ALL CAPITALIZED variable must must be present with the formatting
# present in this file.  These required variables are:
#
# FILL_VALUE, DEFAULT_METADATA, FILE_SPECIFIC_METADATA, FIELD_MAPPINGS,
# DEFAULT_FIELD_NAMES
#
# This section makes generous use of the variables in the Simple Configuration
# section, this is not required, but simplifies and enforces uniformity on
# the configuration.
##############################################################################
##############################################################################


##############################################################################
# Parameters
#
# Various parameters used in Py-ART.
##############################################################################

# the default fill value for masked arrays and the _FillValue key
FILL_VALUE = fill_value

# The DEFAULT_FIELD_NAMES controls the field names which are used in the
# correction and retrieval algorithms in Py-ART. The keys of the dictionary
# are "internal" names which cannot change, the values are the field names
# which will be used in the algorithms by default. For best results use the
# names defined by the variables in simple configuration section which are
# also used in the DEFAULT_METADATA and FIELD_MAPPINGS variable. If you
# choose to change a field name the names should also be changed in the
# DEFAULT_METADATA and FIELD_MAPPINGS variable. This is not required but
# highly suggested.

DEFAULT_FIELD_NAMES = {
    # Internal field name (do not change): field name used (can change)
    'reflectivity': reflectivity,
    'corrected_reflectivity': corrected_reflectivity,
    'total_power': total_power,
    'velocity': velocity,
    'corrected_velocity': corrected_velocity,
    'simulated_velocity': simulated_velocity,
    'spectrum_width': spectrum_width,
    'differential_reflectivity': differential_reflectivity,
    'corrected_differential_reflectivity':
    corrected_differential_reflectivity,
    'cross_correlation_ratio': cross_correlation_ratio,
    'normalized_coherent_power': normalized_coherent_power,
    'differential_phase': differential_phase,
    'unfolded_differential_phase': unfolded_differential_phase,
    'corrected_differential_phase': corrected_differential_phase,
    'specific_differential_phase': specific_differential_phase,
    'corrected_specific_differential_phase':
    corrected_specific_differential_phase,
    'linear_depolarization_ratio': linear_depolarization_ratio,
    'linear_depolarization_ratio_h': linear_depolarization_ratio_h,
    'linear_depolarization_ratio_v': linear_depolarization_ratio_v,
    'signal_to_noise_ratio': signal_to_noise_ratio,
    'rain_rate': rain_rate,
    'radar_estimated_rain_rate': radar_estimated_rain_rate,
    'radar_echo_classification': radar_echo_classification,
    'terrain_altitude': terrain_altitude,
    'bent_terrain_altitude': bent_terrain_altitude,
    'terrain_slope': terrain_slope,
    'terrain_aspect': terrain_aspect,
    'elevation_angle': elevation_angle,
    'visibility': visibility,
    'min_vis_altitude': min_vis_altitude,
    'min_vis_elevation': min_vis_elevation,
    'incident_angle': incident_angle,
    'effective_area': effective_area,
    'sigma_0': sigma_0,
    'rcs_clutter': rcs_clutter,
    'dBm_clutter': dBm_clutter,
    'dBZ_clutter': dBZ_clutter,
    'visibility_polar': visibility_polar,
    'specific_attenuation': specific_attenuation,
    'differential_phase_texture': differential_phase_texture,
    'eastward_wind_component': eastward_wind_component,
    'northward_wind_component': northward_wind_component,
    'vertical_wind_component': vertical_wind_component,
    'height': height,
    'height_over_iso0': height_over_iso0,
    'interpolated_profile': interpolated_profile,
    'path_integrated_attenuation': path_integrated_attenuation,
    'specific_differential_attenuation': specific_differential_attenuation,
    'path_integrated_differential_attenuation':
    path_integrated_differential_attenuation
}


##############################################################################
# Default metadata
#
# The DEFAULT_METADATA dictionary contains dictionaries which provide the
# default radar attribute and field metadata. When reading in a file with
# Py-ART the FILE_SPECIFIC_METADATA variable is first queued for a metadata
# dictionary, if it is not found then the metadata in DEFAULT_METADATA is
# utilized.
##############################################################################

DEFAULT_METADATA = {

    # Metadata for radar attributes. These closely follow the CF/Radial
    # standard
    'azimuth': {
        'units': 'degrees',
        'standard_name': 'beam_azimuth_angle',
        'long_name': 'azimuth_angle_from_true_north',
        'axis': 'radial_azimuth_coordinate',
        'comment': 'Azimuth of antenna relative to true north'},

    'elevation': {
        'units': 'degrees',
        'standard_name': 'beam_elevation_angle',
        'long_name': 'elevation_angle_from_horizontal_plane',
        'axis': 'radial_elevation_coordinate',
        'comment': 'Elevation of antenna relative to the horizontal plane'},

    'scan_rate': {
        'units': 'degrees_per_second',
        'long_name': 'Antenna angle scan rate'},

    'range': {
        'units': 'meters',
        'standard_name': 'projection_range_coordinate',
        'long_name': 'range_to_measurement_volume',
        'axis': 'radial_range_coordinate',
        'spacing_is_constant': 'true',
        'comment': (
            'Coordinate variable for range. Range to center of each bin.')},

    'time': {
        'units': 'seconds',
        'standard_name': 'time',
        'long_name': 'time_in_seconds_since_volume_start',
        'calendar': 'gregorian',
        'comment': ('Coordinate variable for time. '
                    'Time at the center of each ray, in fractional seconds '
                    'since the global variable time_coverage_start')},

    'metadata': {
        'Conventions': 'CF/Radial instrument_parameters',
        'version': '1.3',
        'title': '',
        'institution': '',
        'references': '',
        'source': '',
        'history': '',
        'comment': '',
        'instrument_name': ''},


    # Metadata for radar sweep information dictionaries
    'sweep_number': {
        'units': 'count',
        'standard_name': 'sweep_number',
        'long_name': 'Sweep number'},

    'sweep_mode': {
        'units': 'unitless',
        'standard_name': 'sweep_mode',
        'long_name': 'Sweep mode',
        'comment': ('Options are: "sector", "coplane", "rhi", '
                    '"vertical_pointing", "idle", "azimuth_surveillance", '
                    '"elevation_surveillance", "sunscan", "pointing", '
                    '"manual_ppi", "manual_rhi"')},

    'fixed_angle': {
        'long_name': 'Target angle for sweep',
        'units': 'degrees',
        'standard_name': 'target_fixed_angle'},

    'sweep_start_ray_index': {
        'long_name': 'Index of first ray in sweep, 0-based',
        'units': 'count'},

    'sweep_end_ray_index': {
        'long_name': 'Index of last ray in sweep, 0-based',
        'units': 'count'},

    'rays_per_sweep': {
        'long_name': 'Number of rays in each sweep',
        'units': 'count'},

    'target_scan_rate': {
        'long_name': 'Target scan rate for sweep',
        'units': 'degrees_per_second',
        },

    'rays_are_indexed': {
        'long_name': 'Flag for indexed rays',
        'units': 'unitless',
        'options': ('true: rays are indexed to a regular grid, ' +
                    'false: rays are not indexed to a regular grid'),
        },

    'ray_angle_res': {
        'long_name': 'Angular resolution between rays',
        'units': 'degrees',
        'comment': 'Only applicable when rays_are_indexed variable is true',
        },

    # Metadata for radar location attributes
    'latitude': {
        'long_name': 'Latitude',
        'standard_name': 'Latitude',
        'units': 'degrees_north'},

    'longitude': {
        'long_name': 'Longitude',
        'standard_name': 'Longitude',
        'units': 'degrees_east'},

    'altitude': {
        'long_name': 'Altitude',
        'standard_name': 'Altitude',
        'units': 'meters',
        'positive': 'up'},

    'gate_x': {
        'long_name': 'Cartesian x location of gate with origin at the radar',
        'units': 'meters'},

    'gate_y': {
        'long_name': 'Cartesian y location of gate with origin at the radar',
        'units': 'meters'},

    'gate_z': {
        'long_name': 'Cartesian z location of gate with origin at the radar',
        'units': 'meters'},

    'gate_edge_x': {
        'long_name': 'Cartesian x location of the edges of each gate',
        'units': 'meters'},

    'gate_edge_y': {
        'long_name': 'Cartesian y location of the edges of each gate',
        'units': 'meters'},

    'gate_edge_z': {
        'long_name': 'Cartesian z location of the edges of each gate',
        'units': 'meters'},

    'gate_longitude': {
        'long_name': 'Longitude of radar gate.',
        'units': 'degrees_north'},

    'gate_latitude': {
        'long_name': 'Latitude of radar gate',
        'units': 'degrees_east'},

    'gate_altitude': {
        'long_name': 'Altitude of radar gate',
        'units': 'meters'},

    # Metadata for instrument_parameter dictionary
    'prt_mode': {
        'comments': ('Pulsing mode Options are: "fixed", "staggered", '
                     '"dual". Assumed "fixed" if missing.'),
        'meta_group': 'instrument_parameters',
        'long_name': 'Pulsing mode',
        'units': 'unitless'},

    'nyquist_velocity': {
        'units': 'meters_per_second',
        'comments': "Unambiguous velocity",
        'meta_group': 'instrument_parameters',
        'long_name': 'Nyquist velocity'},

    'prt': {
        'units': 'seconds',
        'comments': ("Pulse repetition time. For staggered prt, "
                     "also see prt_ratio."),
        'meta_group': 'instrument_parameters',
        'long_name': 'Pulse repetition time'},

    'unambiguous_range': {
        'units': 'meters',
        'comments': 'Unambiguous range',
        'meta_group': 'instrument_parameters',
        'long_name': 'Unambiguous range'},

    'pulse_width': {
        'units': 'seconds',
        'comments': 'Pulse width',
        'meta_group': 'instrument_parameters',
        'long_name': 'Pulse width'},

    'prt_ratio': {
        'units': 'unitless',
        'meta_group': 'instrument_parameters',
        'long_name': 'Pulse repetition frequency ratio'},

    'frequency': {
        'units': 's-1',
        'meta_group': 'instrument_parameters',
        'long_name': 'Radiation frequency'},

    'n_samples': {
        'units': 'unitless',
        'meta_group': 'instrument_parameters',
        'long_name': 'Number of samples used to compute moments'},

    # non-standard parameter for specifying the PRF high/low for each ray
    'prf_flag': {
        'units': 'unitless',
        'comments': "PRF used to collect ray. 0 for high PRF, 1 for low PRF.",
        'meta_group': 'instrument_parameters',
        'long_name': 'PRF flag'},

    # Metadata for radar_parameter sub-convention
    'radar_beam_width_h': {
        'units': 'degrees',
        'meta_group': 'radar_parameters',
        'long_name': 'Antenna beam width H polarization'},

    'radar_beam_width_v': {
        'units': 'degrees',
        'meta_group': 'radar_parameters',
        'long_name': 'Antenna beam width V polarization'},

    # Metadata for airborne radar parameters
    'rotation': {
        'units': 'degrees',
        'standard_name': 'ray_rotation_angle_relative_to_platform',
        'long_name': 'Ray rotation angle relative to platform'},

    'tilt': {
        'units': 'degrees',
        'standard_name': 'ray_tilt_angle_relative_to_platform',
        'long_name': 'Ray tilt angle relative to platform'},

    'roll': {
        'units': 'degrees',
        'standard_name': 'platform_roll_angle',
        'long_name': 'Platform roll angle'},

    'drift': {
        'units': 'degrees',
        'standard_name': 'platform_drift_angle',
        'long_name': 'Platform drift angle'},

    'heading': {
        'units': 'degrees',
        'standard_name': 'platform_heading_angle',
        'long_name': 'Platform heading angle'},

    'pitch': {
        'units': 'degrees',
        'standard_name': 'platform_pitch_angle',
        'long_name': 'Platform pitch angle'},

    'georefs_applied': {
        'units': 'unitless',
        'standard_name': 'georefs_have_been_applied_to_ray',
        'long_name': 'Geoferences have been applied to ray',
        'comment': '1 if georefs have been applied, 0 otherwise'},

    # Reflectivity fields
    reflectivity: {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Reflectivity',
        'coordinates': 'elevation azimuth range'},

    corrected_reflectivity: {
        'units': 'dBZ',
        'standard_name': 'corrected_equivalent_reflectivity_factor',
        'long_name': 'Corrected reflectivity',
        'coordinates': 'elevation azimuth range'},

    total_power: {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Total power',
        'coordinates': 'elevation azimuth range'},

    # Velocity fields
    velocity: {
        'units': 'meters_per_second',
        'standard_name': 'radial_velocity_of_scatterers_away_from_instrument',
        'long_name': 'Mean dopper velocity',
        'coordinates': 'elevation azimuth range'},

    corrected_velocity: {
        'units': 'meters_per_second',
        'standard_name': (
            'corrected_radial_velocity_of_scatterers_away_from_instrument'),
        'long_name': 'Corrected mean doppler velocity',
        'coordinates': 'elevation azimuth range'},

    simulated_velocity: {
        'units': 'meters_per_second',
        'standard_name': (
            'radial_velocity_of_scatterers_away_from_instrument'),
        'long_name': 'Simulated mean doppler velocity',
        'coordinates': 'elevation azimuth range'},

    # Spectrum width fields
    spectrum_width: {
        'units': 'meters_per_second',
        'standard_name': 'doppler_spectrum_width',
        'long_name': 'Doppler spectrum width',
        'coordinates': 'elevation azimuth range'},

    # Dual-polarization fields
    differential_reflectivity: {
        'units': 'dB',
        'standard_name': 'log_differential_reflectivity_hv',
        'long_name': 'Differential reflectivity',
        'coordinates': 'elevation azimuth range'},

    corrected_differential_reflectivity: {
        'units': 'dB',
        'standard_name': 'corrected_log_differential_reflectivity_hv',
        'long_name': 'Corrected differential reflectivity',
        'coordinates': 'elevation azimuth range'},

    cross_correlation_ratio: {
        'units': 'ratio',
        'standard_name': 'cross_correlation_ratio_hv',
        'long_name': 'Cross correlation ratio (RHOHV)',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},

    normalized_coherent_power: {
        'units': 'ratio',
        'standard_name': 'normalized_coherent_power',
        'long_name': 'Normalized coherent power',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'comment': 'Also know as signal quality index (SQI)',
        'coordinates': 'elevation azimuth range'},

    differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'Differential phase (PhiDP)',
        'valid_max': 180.0,
        'valid_min': -180.0,
        'coordinates': 'elevation azimuth range'},

    unfolded_differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'Unfolded differential phase',
        'coordinates': 'elevation azimuth range'},

    corrected_differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'Corrected differential phase',
        'coordinates': 'elevation azimuth range'},

    specific_differential_phase: {
        'units': 'degrees/km',
        'standard_name': 'specific_differential_phase_hv',
        'long_name': 'Specific differential phase (KDP)',
        'coordinates': 'elevation azimuth range'},

    corrected_specific_differential_phase: {
        'units': 'degrees/km',
        'standard_name': 'specific_differential_phase_hv',
        'long_name': 'Corrected specific differential phase (KDP)',
        'coordinates': 'elevation azimuth range'},


    # Depolarization ratio fields
    linear_depolarization_ratio: {
        'units': 'dB',
        'standard_name': 'log_linear_depolarization_ratio_hv',
        'long_name': 'Linear depolarization ratio',
        'coordinates': 'elevation azimuth range'},

    linear_depolarization_ratio_h: {
        'units': 'dB',
        'standard_name': 'log_linear_depolarization_ratio_h',
        'long_name': 'Linear depolarization ratio horizontal',
        'coordinates': 'elevation azimuth range'},

    linear_depolarization_ratio_v: {
        'units': 'dB',
        'standard_name': 'log_linear_depolarization_ratio_v',
        'long_name': 'Linear depolarization ratio vertical',
        'coordinates': 'elevation azimuth range'},

    # Misc fields
    signal_to_noise_ratio: {
        'units': 'dB',
        'standard_name': 'signal_to_noise_ratio',
        'long_name': 'Signal to noise ratio',
        'coordinates': 'elevation azimuth range'},

    rain_rate: {
        'units': 'kg/m2/s',
        'standard_name': 'rain_rate',
        'long_name': 'Rain rate',
        'coordinates': 'elevation azimuth range'},

    radar_estimated_rain_rate: {
        'units': 'mm/hr',
        'standard_name': 'radar_estimated_rain_rate',
        'long_name': 'Radar estimated rain rate',
        'coordinates': 'elevation azimuth range'},

    radar_echo_classification: {
        'units': 'legend',
        'standard_name': 'radar_echo_classification',
        'long_name': 'Radar Echo classification',
        'coordinates': 'elevation azimuth range'},

    specific_attenuation: {
        'units': 'dB/km',
        'standard_name': 'specific_attenuation',
        'long_name': 'Specific attenuation',
        'valid_min': 0.0,
        'valid_max': 1.0,
        'coordinates': 'elevation azimuth range'},

    # Visibility processing fields
    terrain_altitude: {'units': 'meters',
          'standard_name': 'terrain_altitude',
          'long_name': 'Altitude',
          'coordinates': 'x y'},

   bent_terrain_altitude: {'units': 'meters',
          'standard_name': 'bent_terrain_altitude',
          'long_name': 'Altitude after refraction correction',
          'coordinates': 'x y'},

    terrain_slope: {'units': 'degrees',
          'standard_name': 'terrain_slope',
          'long_name': 'Slope of the terrain',
          'coordinates': 'x y'},

    terrain_aspect: {'units': 'degrees',
          'standard_name': 'terrain_aspect',
          'long_name': 'Orientation of the terrain',
          'coordinates': 'x y'},

    elevation_angle: {'units': 'degrees',
          'standard_name': 'elevation_angle',
          'long_name': 'Elevation angle at terrain [deg]',
          'coordinates': 'x y'},

    visibility: {'units': '-',
          'standard_name': 'visibility',
          'long_name': 'Radar visibility',
          'coordinates': 'x y'},

    min_vis_altitude: {'units': 'meters',
          'standard_name': 'min_vis_altitude',
          'long_name': 'Minimum visible altitude',
          'coordinates': 'x y'},

    min_vis_elevation: {'units': 'degrees',
          'standard_name': 'min_vis_elevation',
          'long_name': 'Minimum visible elevation angle',
          'coordinates': 'x y'},

    incident_angle: {'units': 'degrees',
          'standard_name': 'incident_angle',
          'long_name': 'Angle of incidence at terrain',
          'coordinates': 'x y'},

    effective_area: {'units': 'm',
          'standard_name': 'effective_area',
          'long_name': 'Effective backscattering area',
          'coordinates': 'x y'},

    sigma_0: {'units': '-',
          'standard_name': 'sigma_0',
          'long_name': 'ratio RCS/backscattering area',
          'coordinates': 'x y'},


    rcs_clutter: {'units': 'square meters',
          'standard_name': 'rcs_clutter',
          'long_name': 'RCS of the ground clutter',
          'coordinates': 'elevation azimuth elevation'},

    dBm_clutter: {'units': 'dBm',
          'standard_name': 'dBm_clutter',
          'long_name': 'Ground clutter power signal',
          'coordinates': 'elevation azimuth elevation'},

    dBZ_clutter: {'units': 'dBZ',
          'standard_name': 'dBZ_clutter',
          'long_name': 'Ground clutter reflectivity',
          'coordinates': 'elevation azimuth elevation'},

    visibility_polar: {'units': '%',
          'standard_name': 'visibility_polar',
          'long_name': 'Visibility',
          'coordinates': 'elevation azimuth range'},

    # Textures
    differential_phase_texture: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv_texture',
        'long_name': 'Texture of differential phase (PhiDP)',
        'coordinates': 'elevation azimuth range'},

    # Wind retrieval fields
    eastward_wind_component: {
        'units': 'meters_per_second',
        'standard_name': 'eastward_wind_component',
        'long_name': 'Eastward wind component'},

    northward_wind_component: {
        'units': 'meters_per_second',
        'standard_name': 'northward_wind_component',
        'long_name': 'Northward wind component'},

    vertical_wind_component: {
        'units': 'meters_per_second',
        'standard_name': 'vertical_wind_component',
        'long_name': 'Vertical wind component'},

    # profile variables
    height: {
        'long_name': 'Height of radar beam',
        'standard_name': 'height',
        'units': 'meters'},

    interpolated_profile: {
        'long_name': 'Interpolated profile',
        'standard_name':  'interpolated_profile',
        'units': 'unknown'},

    # Grid metadata

    'grid_time': {
        'units': 'seconds',
        'standard_name': 'time',
        'long_name': 'Time of grid',
        'calendar': 'gregorian'},

    'origin_longitude': {
        'long_name': 'Longitude at grid origin',
        'units': 'degrees_east',
        'standard_name': 'longitude',
        'valid_min': -180.,
        'valid_max': 180.},

    'origin_latitude': {
        'long_name': 'Latitude at grid origin',
        'units': 'degrees_north',
        'standard_name': 'latitude',
        'valid_min': -90.,
        'valid_max': 90.},

    'origin_altitude': {
        'long_name': 'Altitude at grid origin',
        'units': 'm',
        'standard_name': 'altitude'},

    'x': {
        'standard_name': 'projection_x_coordinate',
        'long_name': 'X distance on the projection plane from the origin',
        'axis': 'X',
        'units': 'm'},

    'y': {
        'standard_name': 'projection_y_coordinate',
        'long_name': 'Y distance on the projection plane from the origin',
        'axis': 'Y',
        'units': 'm'},

    'z': {
        'standard_name': 'projection_z_coordinate',
        'long_name': 'Z distance on the projection plane from the origin',
        'axis': 'Z',
        'units': 'm',
        'positive': 'up'},

    'point_x': {
        'long_name': 'Cartesian x distance of each grid point from the origin',
        'units': 'meters'},

    'point_y': {
        'long_name': 'Cartesian y distance of each grid point from the origin',
        'units': 'meters'},

    'point_z': {
        'long_name': 'Cartesian z distance of each grid point from the origin',
        'positive': 'up',
        'units': 'meters'},

    'point_longitude': {
        'long_name': 'Longitude of each grid point',
        'units': 'degrees_north'},

    'point_latitude': {
        'long_name': 'Latitude of each grid point',
        'units': 'degrees_east'},

    'point_altitude': {
        'long_name': 'Altitude of each grid point',
        'units': 'meters'},

    'radar_latitude': {
        'long_name': 'Latitude of radars used to make the grid.',
        'units': 'degrees_north', },

    'radar_longitude': {
        'long_name': 'Longitude of radars used to make the grid.',
        'units': 'degrees_east', },

    'radar_altitude': {
        'long_name': 'Altitude of radars used to make the grid.',
        'units': 'm', },

    'radar_time': {
        'calendar': 'gregorian',
        'long_name': 'Time in seconds of the volume start for each radar'},

    'radar_name': {
        'long_name': 'Name of radar used to make the grid', },

}


##############################################################################
# File specific metadata
#
# These dictionaries define metadata that is to be used only when reading in
# a given type of file.  This metadata is used in place of the
# DEFAULT_METADATA when it is avialable.  The main use of these variable
# is to define field specific data, it is safe to leave some/all of these
# empty if the default metadata is acceptable.
##############################################################################

# Metadata for Sigmet/IRIS files
sigmet_metadata = {}

# Metadata for NEXRAD Level II files (Archive and CDM files)
nexrad_metadata = {
    reflectivity: {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Reflectivity',
        'valid_max': 94.5,
        'valid_min': -32.0,
        'coordinates': 'elevation azimuth range'},

    velocity: {
        'units': 'meters_per_second',
        'standard_name': 'radial_velocity_of_scatterers_away_from_instrument',
        'long_name': 'Mean doppler Velocity',
        'valid_max': 95.0,
        'valid_min': -95.0,
        'coordinates': 'elevation azimuth range'},

    spectrum_width: {
        'units': 'meters_per_second',
        'standard_name': 'doppler_spectrum_width',
        'long_name': 'Spectrum Width',
        'valid_max': 63.0,
        'valid_min': -63.5,
        'coordinates': 'elevation azimuth range'},

    differential_reflectivity: {
        'units': 'dB',
        'standard_name': 'log_differential_reflectivity_hv',
        'long_name': 'log_differential_reflectivity_hv',
        'valid_max': 7.9375,
        'valid_min': -7.8750,
        'coordinates': 'elevation azimuth range'},

    differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'differential_phase_hv',
        'valid_max': 360.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},

    cross_correlation_ratio: {
        'units': 'ratio',
        'standard_name': 'cross_correlation_ratio_hv',
        'long_name': 'Cross correlation_ratio (RHOHV)',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},
}

# Metadata for NEXRAD Level 3 Products
nexrad_level3_metadata = {

    radar_estimated_rain_rate: {
        'units': 'inches',
        'standard_name': 'radar_estimated_rain_rate',
        'long_name': 'Radar estimated rain rate',
        'coordinates': 'elevation azimuth range'},

    radar_echo_classification: {
        'units': 'legend',
        'standard_name': 'radar_echo_classification',
        'long_name': 'Radar echo classification',
        'options': ('0: Below Threshold (ND), '
                    '10: Biological (BI), '
                    '20: Anomalous Propagation/Group Clutter (GC), '
                    '30: Ice Crystals (IC), '
                    '40: Dry Snow (DS), '
                    '50: Wet Snow (WS), '
                    '60: Light and/or Moderate Rain (RA), '
                    '70: Heavy Rain (HR), '
                    '80: Big Drops (rain) (BD), '
                    '90: Graupel (GR), '
                    '100: Hail, possibly with rain (HA), '
                    '140: Unknown Classification (UK), '
                    '150: Range Folded (RH)'),
        'coordinates': 'elevation azimuth range'},
}

# Metadata for CF/Radial files
cfradial_metadata = {}

# Metadata for MDV files
mdv_metadata = {}

# Metadata for RSL files
rsl_metadata = {}

# Metadata for CSU-CHILL, CHL files
chl_metadata = {}

FILE_SPECIFIC_METADATA = {      # Required
    'sigmet': sigmet_metadata,
    'nexrad_archive': nexrad_metadata,
    'nexrad_cdm': nexrad_metadata,
    'nexrad_level3': nexrad_level3_metadata,
    'cfradial': cfradial_metadata,
    'mdv': mdv_metadata,
    'rsl': rsl_metadata,
    'chl': chl_metadata,
}

##############################################################################
# Field name mapping
#
# These dictionaries map file field names or data types to a radar field
# name.  These are used to populate the radar.fields dictionary during a read
# in Py-ART.  A value of None will not include that field in the radar object.
# These can be over-ridden on a per-read basis using the field_mapping
# parameter, or using setting the file_field_names parameter to True.
##############################################################################

# Sigmet/IRIS file field mapping
# Note that multiple sigmet fields map to the same radar field, if
# more than one of these fields are present the radar field will be
# overwritten with the last sigmet field.
sigmet_field_mapping = {
    # Sigmet data type :field name              # (Data_type) Description
    'XHDR': None,                               # (0) Extended Header
    'DBT': total_power,                         # (1) Total Power
    'DBZ': reflectivity,                        # (2) Reflectivity
    'VEL': velocity,                            # (3) Velocity
    'WIDTH': spectrum_width,                    # (4) Width
    'ZDR': differential_reflectivity,           # (5) Diff. reflectivity
    'DBZC': corrected_reflectivity,             # (7) Corrected reflectivity
    'DBT2': total_power,                        # (8) Total Power
    'DBZ2': reflectivity,                       # (9) Reflectivity
    'VEL2': velocity,                           # (10) Velocity
    'WIDTH2': spectrum_width,                   # (11) Width
    'ZDR2': differential_reflectivity,          # (12) Diff. reflectivity
    'RAINRATE2': radar_estimated_rain_rate,     # (13) Rainfall rate
    'KDP': specific_differential_phase,         # (14) KDP (diff. phase)
    'KDP2': specific_differential_phase,        # (15) KDP (diff. phase)
    'PHIDP': differential_phase,                # (16) PhiDP (diff. phase)
    'VELC': corrected_velocity,                 # (17) Corrected velocity
    'SQI': normalized_coherent_power,           # (18) SQI
    'RHOHV': cross_correlation_ratio,           # (19) RhoHV
    'RHOHV2': cross_correlation_ratio,          # (20) RhoHV
    'DBZC2': corrected_reflectivity,            # (21) Corrected Reflec.
    'VELC2': corrected_velocity,                # (21) Corrected Velocity
    'SQI2': normalized_coherent_power,          # (23) SQI
    'PHIDP2': differential_phase,               # (24) PhiDP (diff. phase)
    'LDRH': linear_depolarization_ratio_h,      # (25) LDR xmt H, rcv V
    'LDRH2': linear_depolarization_ratio_h,     # (26) LDR xmt H, rcv V
    'LDRV': linear_depolarization_ratio_v,      # (27) LDR xmt V, rcv H
    'LDRV2': linear_depolarization_ratio_v,     # (28) LDR xmt V, rcv H
    'HEIGHT': None,                             # (32) Height (1/10 km)
    'VIL2': None,                               # (33) Linear Liquid
    'RAW': None,                                # (34) Raw Data
    'SHEAR': None,                              # (35) Wind Shear
    'DIVERGE2': None,                           # (36) Divergence
    'FLIQUID2': None,                           # (37) Floated liquid
    'USER': None,                               # (38) User type
    'OTHER': None,                              # (39) Unspecified
    'DEFORM2': None,                            # (40) Deformation
    'VVEL2': None,                              # (41) Vertical velocity
    'HVEL2': None,                              # (42) Horizontal velocity
    'HDIR2': None,                              # (43) Horiz. wind direction
    'AXDIL2': None,                             # (44) Axis of dilation
    'TIME2': None,                              # (45) Time in seconds
    'RHOH': None,                               # (46) Rho, xmt H, rcv V
    'RHOH2': None,                              # (47) Rho, xmt H, rcv V
    'RHOV': None,                               # (48) Rho, xmt V, rcv H
    'RHOV2': None,                              # (49) Rho, xmt V, rcv H
    'PHIH': None,                               # (50) Phi, xmt H, rcv V
    'PHIH2': None,                              # (51) Phi, xmt H, rcv V
    'PHIV': None,                               # (52) Phi, xmt V, rcv H
    'PHIV2': None,                              # (53) Phi, xmt V, rcv H
    'USER2': None,                              # (54) User type
    'HCLASS': radar_echo_classification,        # (55) Hydrometeor class
    'HCLASS2': radar_echo_classification,       # (56) Hydrometeor class
    'ZDRC': corrected_differential_reflectivity,
                                                # (57) Corrected diff. refl.
    'ZDRC2': corrected_differential_reflectivity,
                                                # (58) Corrected diff. refl.
    'UNKNOWN_59': None,                         # Unknown field
    'UNKNOWN_60': None,                         # Unknown field
    'UNKNOWN_61': None,                         # Unknown field
    'UNKNOWN_62': None,                         # Unknown field
    'UNKNOWN_63': None,                         # Unknown field
    'UNKNOWN_64': None,                         # Unknown field
    'UNKNOWN_65': None,                         # Unknown field
    'UNKNOWN_66': None,                         # Unknown field
    # there may be more field, add as needed
}


# NEXRAD Level II Archive files
nexrad_archive_field_mapping = {
    # NEXRAD field: radar field name
    'REF': reflectivity,
    'VEL': velocity,
    'SW': spectrum_width,
    'ZDR': differential_reflectivity,
    'PHI': differential_phase,
    'RHO': cross_correlation_ratio
}

# NEXRAD Level II CDM files
nexrad_cdm_field_mapping = {
    # CDM variable name (without _HI): radar field name
    'Reflectivity': reflectivity,
    'RadialVelocity': velocity,
    'SpectrumWidth': spectrum_width,
    'DifferentialReflectivity': differential_reflectivity,
    'DifferentialPhase': differential_phase,
    'CorrelationCoefficient': cross_correlation_ratio
}

# NEXRAD Level 3 Product files.
nexrad_level3_mapping = {
    # Message code : field name         # Product name
    19: reflectivity,                   # Base Reflectivity
    20: reflectivity,                   # Base Reflectivity
    25: velocity,                       # Base Velocity
    27: velocity,                       # Base Velocity
    28: spectrum_width,                 # Base Spectrum Width
    30: spectrum_width,                 # Base Spectrum Width
    32: reflectivity,                   # Digital Hybrid Scan Reflectivity
    34: None,                           # Clutter Filter Control
    56: velocity,                       # Storm Relative Mean Radial Velocity
    78: radar_estimated_rain_rate,      # Surface Rainfall Accum. (1 hr)
    79: radar_estimated_rain_rate,      # Surface Rainfall Accum. (3 hr)
    80: radar_estimated_rain_rate,      # Storm Total Rainfall Accumulation
    94: reflectivity,                   # Base Reflectivity Data Array
    99: velocity,                       # Base Velocity Data Array
    134: None,                          # High Resolution VIL
    135: None,                          # Enhanced Echo Tops
    138: radar_estimated_rain_rate,     # Digital Storm Total Precipitation
    159: differential_reflectivity,     # Digital Differential Reflectivity
    161: cross_correlation_ratio,       # Digital Correlation Coefficient
    163: specific_differential_phase,   # Digital Specific Differential Phase
    165: radar_echo_classification,     # Digital Hydrometeor Classification
    169: radar_estimated_rain_rate,     # One Hour Accumulation
    170: radar_estimated_rain_rate,     # Digital Accumulation Array
    171: radar_estimated_rain_rate,     # Storm Total Accumulation
    172: radar_estimated_rain_rate,     # Digital Storm Total Accumulation
    173: radar_estimated_rain_rate,     # Digital User-Selectable Accum.
    174: radar_estimated_rain_rate,     # Digital 1 hr Diff. Accum.
    175: radar_estimated_rain_rate,     # Digital Storm Total Diff. Accum.
    177: radar_echo_classification,     # Hybrid Hydrometeor Classification
    181: reflectivity,                  # Base Reflectivity
    182: velocity,                      # Base Velocity
    186: reflectivity,                  # Base Reflectivity
}

# MDV files
mdv_field_mapping = {
    # MDV moment: radar field name
    'DBZ_F': reflectivity,
    'VEL_F': velocity,
    'WIDTH_F': spectrum_width,
    'ZDR_F': differential_reflectivity,
    'RHOHV_F': cross_correlation_ratio,
    'NCP_F': normalized_coherent_power,
    'KDP_F': specific_differential_phase,
    'PHIDP_F': differential_phase,
    'VEL_COR': corrected_velocity,
    'PHIDP_UNF': unfolded_differential_phase,
    'KDP_SOB': corrected_specific_differential_phase,
    'DBZ_AC': corrected_reflectivity,
    # repeated integer moments
    'DBZ': reflectivity,
    'VEL': velocity,
    'WIDTH': spectrum_width,
    'ZDR': differential_reflectivity,
    'RHOHV': cross_correlation_ratio,
    'NCP': normalized_coherent_power,
    'KDP': specific_differential_phase,
    'PHIDP': differential_phase,
}

# CF/Radial files
cfradial_field_mapping = {}

# RSL files
# Note that multiple RSL field map to the same radar field, if
# more than one of these fields are present in the RSL data structure
# the radar field will be overwritten with the last field.
rsl_field_mapping = {
    # RSL 2 letter field: radar field           # RSL description
    'DZ': reflectivity,                         # reflectivity
    'VR': velocity,                             # velocity
    'SW': spectrum_width,                       # spectrum width
    'CZ': corrected_reflectivity,               # corrected reflectivity
    'ZT': reflectivity,                         # uncorrected reflectivity
    'DR': differential_reflectivity,            # differential reflectivity
    'LR': differential_reflectivity,            # another diff. reflectivity
    'ZD': differential_reflectivity,            # another diff. reflectivity
    'DM': None,                                 # received power
    'RH': cross_correlation_ratio,              # RhoHV
    'PH': differential_phase,                   # PhiDP
    'XZ': None,                                 # X-band reflectivity
    'CD': corrected_differential_reflectivity,  # Corrected DR.
    'MZ': None,                                 # DZ mask
    'MD': None,                                 # DR Mask
    'ZE': corrected_reflectivity,               # edited reflectivity
    'VE': corrected_velocity,                   # edited velocity
    'KD': specific_differential_phase,          # specific diff. phase
    'TI': None,                                 # TIME (unknown)
    'DX': None,                                 # ???
    'CH': None,                                 # ???
    'AH': None,                                 # ???
    'CV': None,                                 # ???
    'AV': None,                                 # ???
    'SQ': normalized_coherent_power,            # Signal Quality Index (sigmet)
    'VS': None,                                 # Radial Vel. combined
    'VL': None,                                 # Radial Vel. combined
    'VG': None,                                 # Radial Vel. combined
    'VT': None,                                 # Radial Vel. combined
    'NP': normalized_coherent_power,            # Normalized Coherent Power
    'HC': radar_echo_classification,            # Hydroclass
    'VC': None,                                 # Radial Vel. Corrected.
    'V2': None,                                 # Radial Vel cut 2
    'S2': None,                                 # Spectrum width cut 2
    'V3': None,                                 # Radial Vel cut 3
    'S3': None,                                 # Spectrum width cut 3
}

chl_field_mapping = {
    # Chill field name : radar field name
    'Z': reflectivity,
    'V': velocity,
    'W': spectrum_width,
    'ZDR': differential_reflectivity,
    'LDRH': linear_depolarization_ratio_h,
    'LDRV': linear_depolarization_ratio_v,
    b'\xce\xa8 DP'.decode('utf-8'): differential_phase,
    'KDP': specific_differential_phase,
    b'\xcf\x81 HV'.decode('utf-8'): cross_correlation_ratio,
    'NCP': normalized_coherent_power,
    # These fields are not mapped by default
    'H Re(lag 1)': None,    # Real part of lag-1 correlation, H Channel
    'V Re(lag 2)': None,    # Real part of lag-2 correlation, V Channel
    'VAvgQ': None,          # Average Q, V Channel
    'V Im(lag 1)': None,    # Imaginary part of lag-1 correlation, V Channel
    'HAvgQ': None,          # Average Q, H Channel
    'H Im(lag 2)': None,    # Imaginary part of lag-2 correlation, H Channel
    'V lag 0': None,        # Absolute value of lag-0 correlation, V Channel
    'H lag 0': None,        # Absolute value of lag-0 correlation, H Channel
    'H lag 0 cx': None,     # Absolute value of lag-0 cross correlation,
                            # H Channel
    'H Im(lag 1)': None,    # Imaginary part of lag-1 correlation, H Channel
    'H Re(lag 2)': None,    # Real part of lag-2 correlation, H Channel
    'V lag 0 cx': None,     # Absolute value of lag-0 cross correlation,
                            # V Channel
    'V Re(lag 1)': None,    # Real part of lag-1 correlation, V Channel
    'V Im(lag 2)': None,    # Imaginary part of lag-2 correlation, V Channel
    'HV lag 0 I': None,     # Real part of cross channel correlation at lag-0
    'HV lag 0 Q': None,     # Imaginary part of cross channel correlation at
                            # lag-0
    'VAvgI': None,          # Average I, V Channel
    'HAvgI': None,          # Average I, H Channel
    b'\xcf\x81 HCX'.decode('utf-8'): None,   # H Co to Cross Correlation
    b'\xcf\x81 VCX'.decode('utf-8'): None,   # V Co to Cross Correlation
}

# GAMIC HDF5 files
gamic_field_mapping = {
    # Description of GAMIC fields taken from Radx source code file:
    # GamicHdf5RadxFile.cc

    # General notes on field names
    # ----------------------------
    # * An 'U' prefix indicated the moment was corrected from a timeseries
    #   that was not clutter corrected.  Fields without a 'U' prefix are
    #   calculated from a timeseries that has been clutter corrected.
    # * An 'A' prefix indicates that the moment has been corrected
    #   for rainfall attenuation.
    # * 'h' and 'v' endings indicate a the horizontal and vertical channels

    # GAMIC field name: radar field name
    'I': None,          # In phase signal
    'Q': None,          # Quadrature signal

    'Z': corrected_reflectivity,
                        # Reflectivity, corrected for 2nd trip & clutter
    'UZ': reflectivity,
                        # Uncorrected reflectivity
    'Zh': corrected_reflectivity,
    'ZH': corrected_reflectivity,
                        # Corrected reflectivity, horizontal channel
    'Zv': None,
    'ZV': None,         # Corrected reflectivity, vertical channel
    'UZh': reflectivity,
    'UH': reflectivity,
                        # Uncorrected reflectivity, horizontal channel
    'UZv': None,
    'UV': None,         # Uncorrected reflectivity, vertical channel
    'AZh': None,        # Refl., rainfall atten. & clutter corrected, h chan.

    # 'F' in velocity fields indicate folded velocities (no de-aliasing),
    # velocities fields without an F have been unfolded.
    'V': corrected_velocity,
                        # Unfolded velocity from corrected timeseries
    'VF': velocity,     # Folded velocity from corr. t.s.
    'UV': None,         # Unfolded velocity from uncorrected timeseries
    'UVF': None,        # Folded velcity from uncorr. t.s.
    'Vh': corrected_velocity,
    'VH': corrected_velocity,
                        # Velocity from corr. t.s., horizontal channel
    'Vv': None,
    'VV': None,         # Velocity from corr. t.s., vertical channel
    'UVh': None,        # Velocity from uncorr. t.s., horizontal channel
    'UVv': None,        # Velocity from uncorr. t.s., vertical channel
    'VFh': velocity,
                        # Folded velocity, corr. t.s., horizontal channel
    'VFv': None,        # Folded velocity, corr. t.s., vertical channel
    'UnV': None,        # Folded velocity from uncorrected timeseries
    'UnVFh': None,      # Folded velocity, uncorr. t.s., horizontal channel
    'UnVFv': None,      # Folded velocity, uncorr. t.s., vertical channel

    # 'C' in spectral width fields indicate that the field has been
    # corrected for decorrelation causes by antenna rotation
    'W': spectrum_width,
                        # Spectral width from clutter corrected time series.
    'UW': None,         # Spectral width from uncorrected timeseries.
    'CW': None,         # Spec. width, antenna rotation corrected, corr. t.s.
    'UCW': None,        # Spec. width, antenna rotation corrected, uncorr t.s.
    'Wh': spectrum_width,
    'WH': spectrum_width,
                        # Spectral width, corr t.s., horizontal channel
    'Wv': None,
    'WV': None,         # Spectral width, corr t.s., vertical channel
    'UWh': None,        # Spectral width, uncorr t.s., horizontal channel
    'UWv': None,        # Spectral width, uncorr t.s., vertical channel
    'CWh': None,        # Spec. width, antenna rot. corr., horizontal channel
    'CWv': None,        # Spec. width, antenna rot. corr., vertical channel

    'SQI': normalized_coherent_power,
                        # Signal quality index
    'SQIh': normalized_coherent_power,
                        # Signal quality index, horizontal channel
    'SQIv': None,       # Signal quality index, vertical channel

    'CCOR': None,       # Clutter power correction
    'CCORh': None,      # Clutter power correction, horizontal channel
    'CCORv': None,      # Clutter power correction, vertical channel

    'SIGPOW': None,     # Singal Power
    'SNR': None,        # Raw signal to noise ratio
    'SNRh': None,       # Raw signal to noise ratio, horizontal channel
    'SNRv': None,       # Raw signal to noise ration, vertical channel

    'DFT': None,        # Signal spectrum amplitude
    'DFTh': None,       # Signal spectrum amplitude, horizontal channel
    'DFTv': None,       # Signal spectrum amplitude, vertical channel

    'LOG': None,        # Logarithmic amplitude 10*log Isq + Qsq
    'LOGh': None,       # Logarithmic amplitude, horizontal channel
    'LOGv': None,       # Logarithmic amplitude, vertical channel

    'CMAP': None,       # Censor map

    # A '1' ending on differential reflectivity fields indicates that the
    # moment has calculated using a 1st LAG algorithm.
    'ZDR': corrected_differential_reflectivity,
                        # Differential reflectivity from corrected timeseries
    'UZDR': differential_reflectivity,
                        # Diff. refl. from uncorrected timeseries
    'AZDR': None,       # Diff. refl., rainfall atten. corr., corr t.s.
    'ZDR1': None,       # Diff. refl., corr. t.s., 1st LAG algo.
    'UZDR1': None,      # Diff. refl., uncorr. t.s., 1st LAG algo.
    'AZDR1': None,      # Diff. refl., rain. atten. corr., corr. t.s., 1st LAG

    'PHI': corrected_differential_phase,
                        # Differential phase from corrected timeseries
    'PHIDP': corrected_differential_phase,
                        # Differential phase from corrected timeseries
    'UPHIDP': differential_phase,
                        # Diff. phase from uncorrected timeseries.
    'PHIH': None,       # Diff. phase, corr. t.s., horizontal channel
    'UPHIH': None,      # Diff. phase, uncorr t.s., hortizontal channel

    'KDP': specific_differential_phase,
                        # Specific differential phase

    'RHO': cross_correlation_ratio,
                        # Cross correlation coefficient from corrected t.s.
    'RHOHV': cross_correlation_ratio,
                        # Cross correlation coefficient from corrected t.s.
    'URHOHV': None,     # Cross correlation coefficient from uncorr. t.s.
    'RHOH': None,       # Cross corr., corr. t.s., horizontal transmit only
    'URHOH': None,      # Cross corr., uncorr t.s., horizontal transmit only

    'LDR': linear_depolarization_ratio,
                        # Linear depolarization ratio from corr. t.s.
    'ULDR': None,       # Linear depolarization ratio  from uncorr. t.s.
}

# UF field mappings
uf_field_mapping = {
    # UF 2 letter field: radar field
    'DZ': reflectivity,
    'CZ': corrected_reflectivity,
    'ZE': corrected_reflectivity,
    'ZT': total_power,
    'UZ': total_power,
    'VR': velocity,
    'VE': corrected_velocity,
    'SW': spectrum_width,
    'ZD': differential_reflectivity,
    'DR': corrected_differential_reflectivity,
    'CD': corrected_differential_reflectivity,
    'LR': linear_depolarization_ratio,
    'PH': differential_phase,
    'KD': specific_differential_phase,
    'RH': cross_correlation_ratio,
    'SQ': normalized_coherent_power,
    'NP': normalized_coherent_power,
    'HC': radar_echo_classification,
}

# Mapping used when writing UF files
write_uf_mapping = {
    # UF 2 letter field: radar field
    reflectivity: 'DZ',
    corrected_reflectivity: 'CZ',
    total_power: 'ZT',
    velocity: 'VR',
    corrected_velocity: 'VE',
    spectrum_width: 'SW',
    differential_reflectivity: 'ZD',
    corrected_differential_reflectivity: 'DR',
    linear_depolarization_ratio: 'LR',
    differential_phase: 'PH',
    specific_differential_phase: 'KD',
    cross_correlation_ratio: 'RH',
    normalized_coherent_power: 'SQ',
    radar_echo_classification: 'HC',
}

FIELD_MAPPINGS = {                  # Required variable
    'sigmet': sigmet_field_mapping,
    'nexrad_archive': nexrad_archive_field_mapping,
    'nexrad_cdm': nexrad_cdm_field_mapping,
    'nexrad_level3': nexrad_level3_mapping,
    'cfradial': cfradial_field_mapping,
    'mdv': mdv_field_mapping,
    'rsl': rsl_field_mapping,
    'chl': chl_field_mapping,
    'gamic': gamic_field_mapping,
    'uf': uf_field_mapping,
    'write_uf': write_uf_mapping,
}


def velocity_limit(container=None, selection=0):
    import pyart
    if isinstance(container, pyart.core.Radar):
        try:
            if selection >= 0 and selection < container.nsweeps:
                vel = container.get_nyquist_vel(selection,
                                                check_uniform=False)
            else:
                vel = container.get_nyquist_vel(0, check_uniform=False)
            return (-vel, vel)
        except LookupError:
            return (-30., 30.)
    else:
        return (-30., 30.)


def spectrum_width_limit(container=None, selection=0):
    import pyart
    if isinstance(container, pyart.core.Radar):
        try:
            if selection >= 0 and selection < container.nsweeps:
                vel = container.get_nyquist_vel(selection,
                                                check_uniform=False)
            else:
                vel = container.get_nyquist_vel(0, check_uniform=False)
            return (0, vel)
        except LookupError:
            return (0, 30.)
    else:
        return (0, 30.)

# map each field to a colormap

DEFAULT_FIELD_COLORMAP = {
    # field name : colormap
    reflectivity: 'pyart_HomeyerRainbow',
    corrected_reflectivity: 'pyart_HomeyerRainbow',
    total_power: 'pyart_HomeyerRainbow',
    signal_to_noise_ratio: 'pyart_Carbone17',

    velocity: 'pyart_BuDRd18',
    corrected_velocity: 'pyart_BuDRd18',
    simulated_velocity: 'pyart_BuDRd18',
    eastward_wind_component: 'pyart_BuDRd18',
    northward_wind_component: 'pyart_BuDRd18',
    vertical_wind_component: 'pyart_BuDRd18',

    spectrum_width: 'pyart_NWS_SPW',

    normalized_coherent_power: 'pyart_Carbone17',

    differential_reflectivity: 'pyart_RefDiff',
    corrected_differential_reflectivity: 'pyart_RefDiff',

    cross_correlation_ratio: 'pyart_RefDiff',

    differential_phase: 'pyart_Wild25',
    unfolded_differential_phase: 'pyart_Wild25',
    corrected_differential_phase: 'pyart_Wild25',

    specific_differential_phase: 'pyart_Theodore16',
    corrected_specific_differential_phase: 'pyart_Theodore16',

    linear_depolarization_ratio: 'pyart_SCook18',
    linear_depolarization_ratio_h: 'pyart_SCook18',
    linear_depolarization_ratio_v: 'pyart_SCook18',

    rain_rate: 'pyart_RRate11',
    radar_estimated_rain_rate: 'pyart_RRate11',

    radar_echo_classification: 'pyart_LangRainbow12',

    specific_attenuation: 'pyart_Carbone17',

    differential_phase_texture: 'pyart_BlueBrown11',
    height: 'pyart_SCook18',
    interpolated_profile: 'pyart_SCook18',

    terrain_altitude: 'terrain',
    bent_terrain_altitude: 'terrain',
    terrain_slope:  'magma',
    terrain_aspect: 'jet',
    visibility: 'Greys',

    # Additional reflectivity like fields
    'CZ': 'pyart_HomeyerRainbow',
    'DZ': 'pyart_HomeyerRainbow',
    'AZ': 'pyart_HomeyerRainbow',
    'Z': 'pyart_HomeyerRainbow',
    'dbz': 'pyart_HomeyerRainbow',
    'DBZ': 'pyart_HomeyerRainbow',
    'dBZ': 'pyart_HomeyerRainbow',
    'DBZH': 'pyart_HomeyerRainbow',
    'DBZ_S': 'pyart_HomeyerRainbow',
    'DBZ_K': 'pyart_HomeyerRainbow',
    'reflectivity_horizontal': 'pyart_HomeyerRainbow',
    'corr_reflectivity': 'pyart_HomeyerRainbow',
}

# map each field to a limit or a limit function

DEFAULT_FIELD_LIMITS = {
    # field name : limits
    reflectivity: (-30., 75.),
    corrected_reflectivity: (-30., 75.),
    total_power: (-30., 75.),
    signal_to_noise_ratio: (-20, 30.),

    velocity: velocity_limit,
    corrected_velocity: velocity_limit,
    eastward_wind_component: velocity_limit,
    northward_wind_component: velocity_limit,
    vertical_wind_component: velocity_limit,

    spectrum_width: spectrum_width_limit,

    normalized_coherent_power: (0., 1.),

    differential_reflectivity: (-1., 8.),
    corrected_differential_reflectivity: (-1., 8.),

    cross_correlation_ratio: (0.5, 1.05),

    differential_phase: (-180, 180.),
    unfolded_differential_phase: (-360, 360.),
    corrected_differential_phase: (-360, 360.),

    specific_differential_phase: (-2., 5.),
    corrected_specific_differential_phase: (-2., 5.),

    linear_depolarization_ratio: (-40., 0.),
    linear_depolarization_ratio_h: (-40., 0.),
    linear_depolarization_ratio_v: (-40., 0.),

    rain_rate: (0., 50.),
    radar_estimated_rain_rate: (0., 50.),

    radar_echo_classification: (0, 11),

    specific_attenuation: (0., 10.),

    differential_phase_texture: (0, 180.),

    height: (0, 20000),
    interpolated_profile: (0, 10000),

    # Additional reflectivity like fields
    'CZ': (-10., 65.),
    'DZ': (-10., 65.),
    'AZ': (-10., 65.),
    'Z': (-10., 65.),
    'dbz': (-10., 65.),
    'DBZ': (-10., 65.),
    'dBZ': (-10., 65.),
    'DBZH': (-10., 65.),
    'DBZ_S': (-10., 65.),
    'DBZ_K': (-10., 65.),
    'reflectivity_horizontal': (-10., 65.),
    'corr_reflectivity': (-10., 65.),
}



"""
pyart.aux_io.odim_h5
====================

Routines for reading ODIM_H5 files.

.. autosummary::
    :toctree: generated/

    read_odim_grid_h5
    read_odim_h5
    read_odim_vp_h5
    _to_str
    _get_odim_h5_sweep_data
    proj4_to_dict
"""

import datetime
from warnings import warn

import numpy as np
try:
    import h5py
    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False

try:
    import pyproj
    _PYPROJ_AVAILABLE = True
except ImportError:
    _PYPROJ_AVAILABLE = False

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar
from ..core.grid import Grid
from ..exceptions import MissingOptionalDependency


ODIM_H5_FIELD_NAMES = {
    'TH': 'unfiltered_reflectivity',
    'TV': 'unfiltered_reflectivity_vv',
    'DBZH': 'reflectivity',
    'DBZV': 'reflectivity_vv',
    'DBZHC': 'corrected_reflectivity',  # Non standard ODIM
    'DBZVC': 'corrected_reflectivity_vv',  # Non standard ODIM
    'THC': 'corrected_unfiltered_reflectivity',  # Non standard ODIM
    'TVC': 'corrected_unfiltered_reflectivity_vv',  # Non standard ODIM
    'ZBIAS': 'reflectivity_bias',  # Non standard ODIM
    'etah': 'volumetric_reflectivity',  # Non standard ODIM
    'etav': 'volumetric_reflectivity_vv',  # Non standard ODIM
    'RCSH': 'radar_cross_section_hh',  # Non standard ODIM
    'RCSV': 'radar_cross_section_vv',  # Non standard ODIM
    'ZDR': 'differential_reflectivity',
    'ZDRU': 'unfiltered_differential_reflectivity',  # Non standard ODIM
    'ZDRC': 'corrected_differential_reflectivity',  # Non standard ODIM
    'ZDRUC': 'corrected_unfiltered_differential_reflectivity',  # Non standard ODIM
    'ZDRPREC': 'differential_reflectivity_in_precipitation',  # Non standard ODIM
    'ZDRSNOW': 'differential_reflectivity_in_snow',  # Non standard ODIM
    'DBMH': 'signal_power_hh',  # Non standard ODIM
    'DBMV': 'signal_power_vv',  # Non standard ODIM
    'NDBZH': 'noisedBZ_hh',  # Non standard ODIM
    'NDBZV': 'noisedBZ_vv',  # Non standard ODIM
    'DBM_SUNHIT': 'sun_hit_power_h',  # Non standard ODIM
    'DBMV_SUNHIT': 'sun_hit_power_v',  # Non standard ODIM
    'ZDR_SUNHIT': 'sun_hit_differential_reflectivity',  # Non standard ODIM
    'DBM_SUNEST': 'sun_est_power_h',  # Non standard ODIM
    'DBMV_SUNEST': 'sun_est_power_v',  # Non standard ODIM
    'ZDR_SUNEST': 'sun_est_differential_reflectivity',  # Non standard ODIM
    'POSH_SUNHIT': 'sun_hit_h',  # Non standard ODIM
    'POSV_SUNHIT': 'sun_hit_v',  # Non standard ODIM
    'POSZDR_SUNHIT': 'sun_hit_zdr',  # Non standard ODIM
    'RHOHV': 'cross_correlation_ratio',
    'URHOHV': 'uncorrected_cross_correlation_ratio',  # Non standard ODIM
    'RHOHVC': 'corrected_cross_correlation_ratio',  # Non standard ODIM
    'RHOHVRAIN': 'cross_correlation_ratio_in_rain',  # Non standard ODIM
    'LRHOHV': 'logarithmic_cross_correlation_ratio',  # Non standard ODIM
    'CDR': 'circular_depolarization_ratio',  # Non standard ODIM
    'LDR': 'linear_polarization_ratio',
    'PHIDP': 'differential_phase',
    'UPHIDP': 'uncorrected_differential_phase',  # Non standard ODIM
    'PHIDPC': 'corrected_differential_phase',  # Non standard ODIM
    'PHIDP0': 'system_differential_phase',  # Non standard ODIM
    'PHIDP0_BIN': 'first_gate_differential_phase',  # Non standard ODIM
    'KDP': 'specific_differential_phase',
    'KDPC': 'corrected_specific_differential_phase',  # Non standard ODIM
    'SQIH': 'normalized_coherent_power',
    'SQIV': 'normalized_coherent_power_vv',
    'SNRH': 'signal_to_noise_ratio_hh',
    'SNRV': 'signal_to_noise_ratio_vv',
    'CCORH': 'clutter_correction_hh',  # Not used in Pyrad
    'CCORV': 'clutter_correction_vv',  # Not used in Pyrad
    'RATE': 'radar_estimated_rain_rate',
    'URATE': 'uncorrected_rain_rate',  # Not used in Pyrad
    'HI': 'hail_intensity',  # Not used in Pyrad
    'HP': 'hail_probability',  # Not used in Pyrad
    'ACRR': 'accumulated_precipitation',  # Not used in Pyrad
    'HGHT': 'echotop_height',  # Not used in Pyrad
    'VIL': 'vertical_integrated_liquid_water',  # Not used in Pyrad
    'VRAD': 'velocity',  # marked for deprecation in ODIM HDF5 2.2
    'VRADH': 'velocity',
    'VRADV': 'velocity_vv',
    'VRADHC': 'corrected_velocity',  # Non standard ODIM
    'VRADDH': 'dealiased_velocity',
    'VRADDHC': 'dealiased_corrected_velocity',  # Non standard ODIM
    'VRADDV': 'dealiased_velocity_vv',  # radial velocity, vertical polarisation
    'VRADEST': 'retrieved_velocity',  # Non standard ODIM
    'sd_vvp': 'retrieved_velocity_std',  # special vol2bird
    'VDIFF': 'velocity_difference',  # Non standard ODIM
    'WRAD': 'spectrum_width',  # marked for deprecation
    'WRADH': 'spectrum_width',
    'WRADV': 'spectrum_width_vv',
    'WRADHC': 'corrected_spectrum_width',
    'UWND': 'eastward_wind_component',
    'VWND': 'northward_wind_component',
    'AHWND': 'azimuthal_horizontal_wind_component',  # Non standard ODIM
    'w': 'vertical_wind_component',  # Standard for vertical profile
    'RSHR': 'radial_wind_shear',  # Not used in Pyrad
    'ASHR': 'azimuthal_wind_shear',  # Not used in Pyrad
    'CSHR': 'range_azimuthal_wind_shear',  # Not used in Pyrad
    'ESHR': 'elevation_wind_shear',  # Not used in Pyrad
    'OSHR': 'range_elevation_wind_shear',  # Not used in Pyrad
    'HSHR': 'horizontal_wind_shear',  # Not used in Pyrad
    'VSHR': 'vertical_wind_shear',
    'TSHR': 'three_dimensional_shear',  # Not used in Pyrad
    'ff': 'wind_speed',  # Standard for vertical profile
    'dd': 'wind_direction',  # Standard for vertical profile
    'AH': 'specific_attenuation',  # Non standard ODIM
    'AHC': 'corrected_specific_attenuation',  # Non standard ODIM
    'PIA': 'path_integrated_attenuation',  # Non standard ODIM
    'PIAC': 'corrected_path_integrated_attenuation',  # Non standard ODIM
    'ADP': 'specific_differential_attenuation',  # Non standard ODIM
    'ADPC': 'corrected_specific_differential_attenuation',  # Non standard ODIM
    'PIDA': 'path_integrated_differential_attenuation',  # Non standard ODIM
    'PIDAC': 'corrected_path_integrated_differential_attenuation',  # Non standard ODIM
    'TEMP': 'temperature',  # Non standard ODIM
    'ISO0': 'iso0',  # Non standard ODIM
    'HISO0': 'height_over_iso0',  # Non standard ODIM
    'COSMOIND': 'cosmo_index',  # Non standard ODIM
    'HZTIND': 'hzt_index',  # Non standard ODIM
    'ML': 'melting_layer',  # Non standard ODIM
    'VIS': 'visibility',  # Non standard ODIM
    'ECHOID': 'radar_echo_id',  # Non standard ODIM
    'CLT': 'clutter_exit_code',  # Non standard ODIM
    'OCC': 'occurrence',  # Non standard ODIM
    'OCCFREQ': 'frequency_of_occurrence',  # Non standard ODIM
    'BRDR': 'radar_border',  # Not used in Pyrad
    'QIND': 'signal_quality_index',
    'CLASS': 'radar_echo_classification',
    'CELL': 'vol2bird_echo_classification', # Special vol2bird
    'WEATHER': 'vol2bird_weather', # Special vol2bird
    'BACKGROUND': 'vol2bird_background', # Special vol2bird
    'BIOLOGY': 'vol2bird_biology', # Special vol2bird
    'ENTROPY': 'hydroclass_entropy',  # Non standard ODIM
    'propAG': 'proportion_AG',  # Non standard ODIM
    'propCR': 'proportion_CR',  # Non standard ODIM
    'propLR': 'proportion_LR',  # Non standard ODIM
    'propRP': 'proportion_RP',  # Non standard ODIM
    'propRN': 'proportion_RN',  # Non standard ODIM
    'propVI': 'proportion_VI',  # Non standard ODIM
    'propWS': 'proportion_WS',  # Non standard ODIM
    'propMH': 'proportion_MH',  # Non standard ODIM
    'propIH': 'proportion_IH',  # Non standard ODIM
    'TAFLAG': 'time_avg_flag',  # Non standard ODIM
    'COLGATES': 'colocated_gates',  # Non standard ODIM
    'ns': 'number_of_samples',  # Non standard ODIM
    'dens': 'bird_density',  # standard vol2bird
    'STD': 'standard_deviation',  # Non standard ODIM
    'SUM': 'sum',  # Non standard ODIM
    'SUM2': 'sum_squared',  # Non standard ODIM
    'width': 'height_resolution',  # Special vol2bird
    'gap': 'gap',  # Special vol2bird
    'eta': 'bird_reflectivity',  # Special vol2bird
    'dbz': 'volumetric_reflectivity',  # Special vol2bird
    'n': 'number_of_samples_velocity',  # Special vol2bird
    'n_dbz': 'number_of_samples_reflectivity',  # Special vol2bird
    'n_all': 'number_of_samples_velocity_all',  # Special vol2bird
    'n_dbz_all': 'number_of_samples_reflectivity_all',   # Special vol2bird
    'u': 'eastward_wind_component',   # Special vol2bird
    'v': 'northward_wind_component',   # Special vol2bird
    'CHBZC': 'probability_of_hail',
    'CHMZC': 'maximum_expected_severe_hail_size',
    'CHRZC': 'radar_estimated_rain_rate',  # RZC grid product
    'CHRZF': 'radar_estimated_rain_rate',
    'CHTZC': 'radar_estimated_rain_rate',
    'CHCPC': 'radar_estimated_rain_rate',
    'CHCPCH': 'radar_estimated_rain_rate',
    'CHRFQ': 'radar_estimated_rain_rate',
    'CHAZC*': 'rainfall_accumulation',
    'CHDV*': 'dealiased_velocity',
    'CHOZC': 'reflectivity',
    'CHEZC_015': 'echo_top_15dBZ',
    'CHEZC_020': 'echo_top_20dBZ',
    'CHEZC_045': 'echo_top_45dBZ',
    'CHEZC_050': 'echo_top_50dBZ',
    'CHCZC': 'maximum_echo',
    'CHLZC': 'vertically_integrated_liquid',
}


def read_odim_grid_h5(filename, field_names=None, additional_metadata=None,
                      file_field_names=False, exclude_fields=None,
                      include_fields=None, **kwargs):
    """
    Read a ODIM_H5 gird file.

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.


    Returns
    -------
    grid : Grid
        Grid object containing data from ODIM_H5 file.

    """
    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_odim_h5 but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = ODIM_H5_FIELD_NAMES
    filemetadata = FileMetadata('odim_h5', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    with h5py.File(filename, 'r') as hfile:
        odim_object = _to_str(hfile['what'].attrs['object'])
        if odim_object not in ['COMP', 'CVOL']:
            raise NotImplementedError(
                'object: %s not implemented.' % (odim_object))

        # determine the number of sweeps by the number of groups which
        # begin with dataset
        datasets = [k for k in hfile if k.startswith('dataset')]
        datasets.sort(key=lambda x: int(x[7:]))

        # latitude, longitude and altitude
        x = filemetadata('x')
        y = filemetadata('y')
        z = filemetadata('z')

        h_where = hfile['where'].attrs

        projection = h_where['projdef']

        xvec = np.linspace(
            h_where['UR_lat'], h_where['LL_lat'], h_where['ysize'])
        yvec = np.linspace(
            h_where['LL_lon'], h_where['UR_lon'], h_where['xsize'])

        if _PYPROJ_AVAILABLE:
            wgs84 = pyproj.Proj(4326)
            try:  # pyproj doens't like bytearrays
                projection = projection.decode('utf-8')
            except:
                pass

            coordTrans = pyproj.Transformer.from_proj(wgs84, projection)
            ystart, xstart = coordTrans.transform(
                h_where['LL_lat'], h_where['LL_lon'])
            yend, xend = coordTrans.transform(
                h_where['UR_lat'], h_where['UR_lon'])
            xvec = np.linspace(xstart, xend, h_where['ysize'])
            yvec = np.linspace(ystart, yend, h_where['ysize'])

        x['data'] = xvec
        y['data'] = yvec
        z['data'] = np.array([0], dtype='float64')

        # metadata
        metadata = filemetadata('metadata')
        metadata['source'] = _to_str(hfile['what'].attrs['source'])
        metadata['original_container'] = 'odim_h5'
        metadata['odim_conventions'] = _to_str(hfile.attrs['Conventions'])

        h_what = hfile['what'].attrs
        metadata['version'] = _to_str(h_what['version'])
        metadata['source'] = _to_str(h_what['source'])

        # Get the MeteoSwiss-specific data
        try:
            h_how2 = hfile['how']['MeteoSwiss'].attrs
        except:
            # if no how group exists mock it with an empty dictionary
            h_how2 = {}
        if 'radar' in h_how2:
            metadata['radar'] = h_how2['radar']

        try:
            ds1_how = hfile[datasets[0]]['how'].attrs
        except KeyError:
            # if no how group exists mock it with an empty dictionary
            ds1_how = {}
        if 'system' in ds1_how:
            metadata['system'] = ds1_how['system']
        if 'software' in ds1_how:
            metadata['software'] = ds1_how['software']
        if 'sw_version' in ds1_how:
            metadata['sw_version'] = ds1_how['sw_version']

        dset = 'dataset1'  # we assume only one dataset
        fields = {}
        h_field_keys = [k for k in hfile['dataset1'] if k.startswith('data')]

        if odim_object == 'CVOL':  # CAPPI case
            z['data'] += hfile[dset]['what'].attrs['prodpar']

        odim_fields = [hfile['dataset1']['what'].attrs['prodname']]

        for odim_field, h_field_key in zip(odim_fields, h_field_keys):
            field_name = filemetadata.get_field_name(_to_str(odim_field))

            if field_name is None:
                continue
            if 'what' not in hfile[dset][h_field_key].keys():
                # This is specific to MCH Cartesian products
                gain = hfile[dset]['what'].attrs['gain']
                offset = hfile[dset]['what'].attrs['offset']
                nodata = hfile[dset]['what'].attrs['nodata']

                fdata = _get_odim_h5_sweep_data(hfile[dset][h_field_key],
                                                offset, gain, nodata)
            else:
                fdata = _get_odim_h5_sweep_data(hfile[dset][h_field_key])

            field_dic = filemetadata(field_name)
            field_dic['data'] = fdata
            field_dic['_FillValue'] = get_fillvalue()
            # Keep track of this information to later write correctly ODIM
            field_dic['prodname'] = hfile['dataset1']['what'].attrs['prodname']
            field_dic['product'] = hfile['dataset1']['what'].attrs['product']
            if odim_object == 'CVOL':  # add height info
                field_dic['product'] += '_{:f}'.format(
                    hfile[dset]['what'].attrs['prodpar']).encode('utf-8')
                field_dic['product'] = np.bytes_(field_dic['product'])

            fields[field_name] = field_dic

        _time = filemetadata('time')
        origin_latitude = filemetadata('origin_latitude')
        origin_longitude = filemetadata('origin_longitude')
        origin_altitude = filemetadata('origin_altitude')

        start_date = hfile['dataset1']['what'].attrs['startdate']
        start_time = hfile['dataset1']['what'].attrs['starttime']
        start_time = datetime.datetime.strptime(
            _to_str(start_date + start_time), '%Y%m%d%H%M%S')
        end_date = hfile['dataset1']['what'].attrs['startdate']
        end_time = hfile['dataset1']['what'].attrs['starttime']
        end_time = datetime.datetime.strptime(
            _to_str(end_date + end_time), '%Y%m%d%H%M%S')
        _time['units'] = make_time_unit_str(start_time)
        _time['data'] = [0]

        projection = proj4_to_dict(projection)
        origin_latitude['data'] = np.array([projection['lat_0']])
        origin_longitude['data'] = np.array([projection['lon_0']])
        origin_altitude['data'] = np.array([0.])

        # radar variables
        radar_latitude = None
        radar_longitude = None
        radar_altitude = None
        radar_name = None
        radar_time = None

        return Grid(
            _time, fields, metadata,
            origin_latitude, origin_longitude, origin_altitude, x, y, z,
            projection=projection,
            radar_latitude=radar_latitude, radar_longitude=radar_longitude,
            radar_altitude=radar_altitude, radar_name=radar_name,
            radar_time=radar_time)


def read_odim_h5(filename, field_names=None, additional_metadata=None,
                 file_field_names=False, exclude_fields=None,
                 include_fields=None, **kwargs):
    """
    Read a ODIM_H5 file.

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.


    Returns
    -------
    radar : Radar
        Radar object containing data from ODIM_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_odim_h5 but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = ODIM_H5_FIELD_NAMES
    filemetadata = FileMetadata('odim_h5', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # open the file
    with h5py.File(filename, 'r') as hfile:
        odim_object = _to_str(hfile['what'].attrs['object'])
        if odim_object not in ['PVOL', 'SCAN', 'ELEV', 'AZIM']:
            raise NotImplementedError(
                'object: %s not implemented.' % (odim_object))

        # determine the number of sweeps by the number of groups which
        # begin with dataset
        datasets = [k for k in hfile if k.startswith('dataset')]
        datasets.sort(key=lambda x: int(x[7:]))
        nsweeps = len(datasets)

        # latitude, longitude and altitude
        latitude = filemetadata('latitude')
        longitude = filemetadata('longitude')
        altitude = filemetadata('altitude')

        h_where = hfile['where'].attrs
        latitude['data'] = np.array([h_where['lat']], dtype='float64')
        longitude['data'] = np.array([h_where['lon']], dtype='float64')
        altitude['data'] = np.array([h_where['height']], dtype='float64')

        # metadata
        metadata = filemetadata('metadata')
        metadata['source'] = _to_str(hfile['what'].attrs['source'])
        metadata['original_container'] = 'odim_h5'
        metadata['odim_conventions'] = _to_str(hfile.attrs['Conventions'])

        h_what = hfile['what'].attrs
        metadata['version'] = _to_str(h_what['version'])
        metadata['source'] = _to_str(h_what['source'])

        try:
            ds1_how = hfile[datasets[0]]['how'].attrs
        except KeyError:
            # if no how group exists mock it with an empty dictionary
            ds1_how = {}
        if 'system' in ds1_how:
            metadata['system'] = ds1_how['system']
        if 'software' in ds1_how:
            metadata['software'] = ds1_how['software']
        if 'sw_version' in ds1_how:
            metadata['sw_version'] = ds1_how['sw_version']

        # sweep_start_ray_index, sweep_end_ray_index
        sweep_start_ray_index = filemetadata('sweep_start_ray_index')
        sweep_end_ray_index = filemetadata('sweep_end_ray_index')

        if odim_object in ['AZIM', 'SCAN', 'PVOL']:
            rays_per_sweep = [
                int(hfile[d]['where'].attrs['nrays']) for d in datasets]
        elif odim_object == 'ELEV':
            rays_per_sweep = [
                int(hfile[d]['where'].attrs['angles'].size) for d in datasets]
        total_rays = sum(rays_per_sweep)
        ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
        seri = np.cumsum(rays_per_sweep).astype('int32') - 1
        sweep_start_ray_index['data'] = ssri
        sweep_end_ray_index['data'] = seri

        # sweep_number
        sweep_number = filemetadata('sweep_number')
        sweep_number['data'] = np.arange(nsweeps, dtype='int32')

        # sweep_mode
        sweep_mode = filemetadata('sweep_mode')
        sweep_mode['data'] = np.array(nsweeps * ['azimuth_surveillance'])

        # scan_type
        if odim_object == 'ELEV':
            scan_type = 'rhi'
        else:
            scan_type = 'ppi'

        # fixed_angle
        fixed_angle = filemetadata('fixed_angle')
        if odim_object == 'ELEV':
            sweep_el = [hfile[d]['where'].attrs['az_angle'] for d in datasets]
        else:
            sweep_el = [hfile[d]['where'].attrs['elangle'] for d in datasets]
        fixed_angle['data'] = np.array(sweep_el, dtype='float32')

        # elevation
        elevation = filemetadata('elevation')
        if 'elangles' in ds1_how:
            edata = np.empty(total_rays, dtype='float32')
            for d, start, stop in zip(datasets, ssri, seri):
                edata[start:stop+1] = hfile[d]['how'].attrs['elangles'][:]
            elevation['data'] = edata
        elif odim_object == 'ELEV':
            edata = np.empty(total_rays, dtype='float32')
            for d, start, stop in zip(datasets, ssri, seri):
                edata[start:stop+1] = hfile[d]['where'].attrs['angles'][:]
            elevation['data'] = edata
        else:
            elevation['data'] = np.repeat(sweep_el, rays_per_sweep)

        # range
        _range = filemetadata('range')
        if 'rstart' in hfile['dataset1/where'].attrs:
            # derive range from rstart and rscale attributes if available

            # check that the gate spacing is constant between sweeps
            rstart = [hfile[d]['where'].attrs['rstart'] for d in datasets]
            if any(rstart != rstart[0]):
                raise ValueError('range start changes between sweeps')
            rscale = [hfile[d]['where'].attrs['rscale'] for d in datasets]
            if any(rscale != rscale[0]):
                raise ValueError('range scale changes between sweeps')
            all_sweeps_nbins = [
                hfile[d]['where'].attrs['nbins'] for d in datasets]
            # check for max range off all sweeps
            max_nbins = max(all_sweeps_nbins)

            if isinstance(max_nbins, np.ndarray):
                max_nbins = max_nbins[0]
            else:
                max_nbins = max(all_sweeps_nbins)

            rscenter = 1e3 * rstart[0] + rscale[0] / 2
            _range['data'] = np.arange(rscenter,
                                       rscenter + max_nbins * rscale[0],
                                       rscale[0], dtype='float32')
            _range['meters_to_center_of_first_gate'] = rstart[0] * 1000.
            _range['meters_between_gates'] = float(rscale[0])
        else:
            # if not defined use range attribute which defines the maximum
            # range in km. There is no information on the starting location of
            # the range bins so we assume this to be 0.
            # This most often occurs in RHI files, which technically do not
            # meet the ODIM 2.2 specs. Section 7.4 requires that these files
            # include the where/rstart, where/rscale and where/nbins
            # attributes.
            max_range = [hfile[d]['where'].attrs['range'] for d in datasets]
            if any(max_range != max_range[0]):
                raise ValueError('maximum range changes between sweeps')
            # nbins is required
            max_nbins = hfile['dataset1/data1/data'].shape[1]
            _range['data'] = np.linspace(
                0, max_range[0] * 1000., max_nbins).astype('float32')
            _range['meters_to_center_of_first_gate'] = 0
            _range['meters_between_gates'] = max_range[0] * 1000. / max_nbins

        # azimuth
        azimuth = filemetadata('azimuth')
        az_data = np.ones((total_rays, ), dtype='float32')
        for dset, start, stop in zip(datasets, ssri, seri):
            if odim_object == 'ELEV':
                # all azimuth angles are the sweep azimuth angle
                sweep_az = hfile[dset]['where'].attrs['az_angle']
            elif odim_object == 'AZIM':
                # Sector azimuths are specified in the startaz and stopaz
                # attribute of dataset/where.
                # Assume that the azimuth angles do not pass through 0/360 deg.
                startaz = hfile[dset]['where'].attrs['startaz']
                stopaz = hfile[dset]['where'].attrs['stopaz']
                nrays = stop - start + 1
                sweep_az = np.linspace(startaz, stopaz, nrays, endpoint=True)
            elif ('startazA' in ds1_how) and ('stopazA' in ds1_how):
                # average between start and stop azimuth angles
                startaz = hfile[dset]['how'].attrs['startazA']
                stopaz = hfile[dset]['how'].attrs['stopazA']
                sweep_az = np.angle(
                    (np.exp(1.j*np.deg2rad(startaz)) +
                     np.exp(1.j*np.deg2rad(stopaz))) / 2., deg=True)
                sweep_az[sweep_az < 0.] = 360 + sweep_az[sweep_az < 0.]
            else:
                # according to section 5.1 the first ray points to the
                # northernmost direction and proceeds clockwise for a complete
                # 360 rotation.
                try:
                    astart = hfile[dset]['how'].attrs['astart']
                except KeyError:
                    astart = 0.0
                nrays = hfile[dset]['where'].attrs['nrays']
                da = 360.0 / nrays
                sweep_az = np.arange(astart+da/2., 360., da, dtype='float32')
            az_data[start:stop+1] = sweep_az
        azimuth['data'] = az_data

        # time
        _time = filemetadata('time')
        if ('startazT' in ds1_how) and ('stopazT' in ds1_how):
            # average between startazT and stopazT
            t_data = np.empty((total_rays, ), dtype='float32')
            for dset, start, stop in zip(datasets, ssri, seri):
                t_start = hfile[dset]['how'].attrs['startazT']
                t_stop = hfile[dset]['how'].attrs['stopazT']
                t_data[start:stop+1] = (t_start + t_stop) / 2
            start_epoch = t_data.min()
            start_time = datetime.datetime.utcfromtimestamp(start_epoch)
            _time['units'] = make_time_unit_str(start_time)
            _time['data'] = t_data - start_epoch
        else:
            t_data = np.empty((total_rays, ), dtype='int32')
            # interpolate between each sweep starting and ending time
            for dset, start, stop in zip(datasets, ssri, seri):
                dset_what = hfile[dset]['what'].attrs
                start_str = _to_str(
                    dset_what['startdate'] + dset_what['starttime'])
                end_str = _to_str(dset_what['enddate'] + dset_what['endtime'])
                start_dt = datetime.datetime.strptime(
                    start_str, '%Y%m%d%H%M%S')
                end_dt = datetime.datetime.strptime(end_str, '%Y%m%d%H%M%S')

                time_delta = end_dt - start_dt
                delta_seconds = time_delta.seconds+time_delta.days*3600*24
                rays = stop - start + 1
                sweep_start_epoch = (
                    start_dt - datetime.datetime(1970, 1, 1)).total_seconds()
                t_data[start:stop+1] = (sweep_start_epoch +
                                        np.linspace(0, delta_seconds, rays))
            start_epoch = t_data.min()
            start_time = datetime.datetime.utcfromtimestamp(start_epoch)
            _time['units'] = make_time_unit_str(start_time)
            _time['data'] = (t_data - start_epoch).astype('float32')

        # fields
        fields = {}
        h_field_keys = [k for k in hfile['dataset1'] if k.startswith('data')]
        odim_fields = [hfile['dataset1'][d]['what'].attrs['quantity'] for d in
                       h_field_keys]
        for odim_field, h_field_key in zip(odim_fields, h_field_keys):
            field_name = filemetadata.get_field_name(_to_str(odim_field))
            if field_name is None:
                continue
            fdata = np.ma.zeros((total_rays, max_nbins), dtype='float32')
            start = 0
            # loop on the sweeps, copy data into correct location in data array
            for dset, rays_in_sweep in zip(datasets, rays_per_sweep):
                if h_field_key not in hfile[dset]:
                    warn('{} not in {} in {}'.format(odim_field, h_field_key, dset))
                    continue
                sweep_data = _get_odim_h5_sweep_data(hfile[dset][h_field_key])
                sweep_nbins = sweep_data.shape[1]
                fdata[start:start + rays_in_sweep, :sweep_nbins] = (
                    sweep_data[:])
                # set data to NaN if its beyond the range of this sweep
                fdata[start:start + rays_in_sweep,
                      sweep_nbins:max_nbins] = np.nan
                start += rays_in_sweep
            # create field dictionary
            field_dic = filemetadata(field_name)
            field_dic['data'] = fdata
            field_dic['_FillValue'] = get_fillvalue()
            fields[field_name] = field_dic

    # instrument_parameters
    instrument_parameters = None
    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def read_odim_vp_h5(filename, field_names=None, additional_metadata=None,
                    file_field_names=False, exclude_fields=None,
                    include_fields=None, **kwargs):
    """
    Read a vertical profile ODIM_H5 file like those used by vol2bird

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.


    Returns
    -------
    radar : Radar
        Radar object containing data from ODIM_H5 file.

    """
    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_odim_h5 but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = ODIM_H5_FIELD_NAMES
    filemetadata = FileMetadata('odim_h5', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # open the file
    with h5py.File(filename, 'r') as hfile:
        odim_object = _to_str(hfile['what'].attrs['object'])
        if odim_object != 'VP':
            raise NotImplementedError(
                'object: %s not implemented.' % (odim_object))

        # determine the number of sweeps by the number of groups which
        # begin with dataset
        datasets = [k for k in hfile if k.startswith('dataset')]
        datasets.sort(key=lambda x: int(x[7:]))
        nsweeps = len(datasets)

        # latitude, longitude and altitude
        latitude = filemetadata('latitude')
        longitude = filemetadata('longitude')
        altitude = filemetadata('altitude')

        h_where = hfile['where'].attrs
        latitude['data'] = np.array([h_where['lat']], dtype='float64')
        longitude['data'] = np.array([h_where['lon']], dtype='float64')
        altitude['data'] = np.array([h_where['height']], dtype='float64')

        # metadata
        metadata = filemetadata('metadata')
        metadata['original_container'] = 'odim_h5'
        metadata['odim_conventions'] = _to_str(hfile.attrs['Conventions'])

        h_what = hfile['what'].attrs
        metadata['version'] = _to_str(h_what['version'])
        metadata['source'] = _to_str(h_what['source'])

        try:
            ds1_how = hfile[datasets[0]]['how'].attrs
        except KeyError:
            # if no how group exists mock it with an empty dictionary
            ds1_how = {}
        if 'system' in ds1_how:
            metadata['system'] = ds1_how['system']
        if 'software' in ds1_how:
            metadata['software'] = ds1_how['software']
        if 'sw_version' in ds1_how:
            metadata['sw_version'] = ds1_how['sw_version']

        # sweep_start_ray_index, sweep_end_ray_index
        sweep_start_ray_index = filemetadata('sweep_start_ray_index')
        sweep_end_ray_index = filemetadata('sweep_end_ray_index')

        rays_per_sweep = np.array([1])
        total_rays = sum(rays_per_sweep)
        ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
        seri = np.cumsum(rays_per_sweep).astype('int32') - 1
        sweep_start_ray_index['data'] = ssri
        sweep_end_ray_index['data'] = seri

        # sweep_number
        sweep_number = filemetadata('sweep_number')
        sweep_number['data'] = np.arange(nsweeps, dtype='int32')

        # sweep_mode
        sweep_mode = filemetadata('sweep_mode')
        sweep_mode['data'] = np.array(nsweeps * ['vertical_pointing'])

        # scan_type
        scan_type = 'vertical_pointing'

        # fixed_angle
        fixed_angle = filemetadata('fixed_angle')
        fixed_angle['data'] = np.array([90.], dtype='float32')

        # elevation
        elevation = filemetadata('elevation')
        elevation['data'] = np.array([90.], dtype='float32')

        # range
        _range = filemetadata('range')
        rstart = hfile['where'].attrs['minheight']
        rscale = hfile['where'].attrs['interval']
        rend = hfile['where'].attrs['maxheight']

        rscenter = rstart + rscale / 2
        _range['data'] = np.arange(
            rscenter, rscenter + rend, rscale, dtype='float32')
        _range['meters_to_center_of_first_gate'] = rstart
        _range['meters_between_gates'] = rscale
        max_nbins = _range['data'].size

        # azimuth
        azimuth = filemetadata('azimuth')
        azimuth['data'] = np.array([0.], dtype='float32')

        # time
        _time = filemetadata('time')
        _time['units'] = make_time_unit_str(datetime.datetime.strptime(
            hfile['what'].attrs['date'].decode("utf-8")
            + hfile['what'].attrs['time'].decode("utf-8"), '%Y%m%d%H%M%S'))
        _time['data'] = np.array([0], dtype='float32')

        # fields
        fields = {}
        h_field_keys = [k for k in hfile['dataset1'] if k.startswith('data')]
        odim_fields = [hfile['dataset1'][d]['what'].attrs['quantity'] for d in
                       h_field_keys]
        for odim_field, h_field_key in zip(odim_fields, h_field_keys):
            field_name = filemetadata.get_field_name(_to_str(odim_field))
            if field_name is None:
                continue
            fdata = np.ma.zeros((total_rays, max_nbins), dtype='float32')
            start = 0
            # loop on the sweeps, copy data into correct location in data array
            for dset, rays_in_sweep in zip(datasets, rays_per_sweep):
                sweep_data = _get_odim_h5_sweep_data(hfile[dset][h_field_key])
                sweep_nbins = sweep_data.shape[0]
                fdata[start:start + rays_in_sweep, :sweep_nbins] = (
                    sweep_data.T)
                # set data to NaN if its beyond the range of this sweep
                fdata[start:start + rays_in_sweep,
                      sweep_nbins:max_nbins] = np.nan
                start += rays_in_sweep
            # create field dictionary
            field_dic = filemetadata(field_name)
            field_dic['data'] = fdata
            field_dic['_FillValue'] = get_fillvalue()
            fields[field_name] = field_dic

    # instrument_parameters
    instrument_parameters = None
    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def _to_str(text):
    """ Convert bytes to str if necessary. """
    if hasattr(text, 'decode'):
        return text.decode('utf-8')

    return text


def _get_odim_h5_sweep_data(group, offset=0, gain=1, nodata=np.nan):
    """ Get ODIM_H5 sweet data from an HDF5 group. """

    # mask raw data
    raw_data = group['data'][:]

    try:
        what = group['what']

        if 'nodata' in what.attrs:
            nodata = what.attrs.get('nodata')
            data = np.ma.masked_equal(raw_data, nodata)
        else:
            data = np.ma.masked_array(raw_data)
        if 'undetect' in what.attrs:
            undetect = what.attrs.get('undetect')
            data[data == undetect] = np.ma.masked

        if 'offset' in what.attrs:
            offset = what.attrs.get('offset')
        if 'gain' in what.attrs:
            gain = what.attrs.get('gain')
    except KeyError:
        data = np.ma.masked_equal(raw_data, nodata)

    return data * gain + offset


def proj4_to_dict(proj4str):
    """ Convert proj4 string to dict format"""
    if not isinstance(proj4str, str):
        proj4str = proj4str.decode('utf-8')

    proj4dict = {}
    splitspace = proj4str.split(' ')
    for s in splitspace:
        ssplit = s.split('=')
        key = ssplit[0][1:]
        if key == "no_defs":
            proj4dict[key] = True
        else:
            val = ssplit[1]
            try:
                proj4dict[key] = float(val)
            except:
                proj4dict[key] = val

    return proj4dict



"""
====================================================================
Create an RHI plot with reflectivity contour lines from an MDV file
====================================================================

An example which creates an RHI plot of velocity using a RadarDisplay object
and adding Reflectivity contours from the same MDV file.

"""
print(__doc__)

# Author: Cory Weber (cweber@anl.gov)
# License: BSD 3 clause
import matplotlib.pyplot as plt
import pyart
import numpy as np
import scipy.ndimage as spyi

filename = '034142.mdv'

# create the plot using RadarDisplay
sweep = 2
# read file
radar = pyart.io.read_mdv(filename)
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure(figsize=[20, 5])
ax = fig.add_subplot(111)

# plot velocity
# cmap is the color ramp being used in this case blue to red no 18
# https://github.com/ARM-DOE/pyart/blob/master/pyart/graph/cm.py
# for more information

display.plot('velocity', sweep=sweep, vmin=-20, vmax=20.0, fig=fig,
             ax=ax, cmap='pyart_BuDRd18', colorbar_label='Velocity (m/s)')

# line commented out to show reflectivity
# display.plot('reflectivity', sweep=sweep, vmin=-0, vmax=45.0, fig=fig,ax=ax)

# get data
start = radar.get_start(sweep)
end = radar.get_end(sweep) + 1
data = radar.get_field(sweep, 'reflectivity')
x, y, z = radar.get_gate_x_y_z(sweep, edges=False)

x /= 1000.0
y /= 1000.0
z /= 1000.0

# smooth out the lines
data = spyi.gaussian_filter(data, sigma=1.2)

# calculate (R)ange
R = np.sqrt(x ** 2 + y ** 2) * np.sign(y)
R = -R
display.set_limits(xlim=[25, 0], ylim=[0, 5])

# add contours
# creates steps 35 to 100 by 5
levels = np.arange(35, 100, 5)
# adds coutours to plot
contours = ax.contour(R, z, data, levels, linewidths=1.5, colors='k',
                      linestyles='solid', antialiased=True)

# adds contour labels (fmt= '%r' displays 10.0 vs 10.0000)
plt.clabel(contours, levels, fmt='%r', inline=True, fontsize=10)


# format plot
# add grid (dotted lines, major axis only)
ax.grid(color='k', linestyle=':', linewidth=1, which='major')

# horizontal
ax.axhline(0.9, 0, 1, linestyle='solid', color='k', linewidth=2)
ax.axhline(1.3, 0, 1, linestyle='dashed', color='k', linewidth=2)

# vertical
ax.axvline(15, 0, 1, linestyle='solid', color='#00b4ff', linewidth=2)
ax.axvline(4.5, 0, 1, linestyle='solid', color='#ff6800', linewidth=2)

# setting matplotlib overrides display.plot defaults
ax.set_ylabel('Altitude above CP-2 (km)')

plt.show()



"""
================================================
Bridging to other toolkits (:mod:`pyart.bridge`)
================================================

.. currentmodule:: pyart.bridge

Py-ART can act as bridge to other community software projects.

The functionality in this namespace is available in other pyart namespaces.

Phase functions
===============

.. autosummary::
    :toctree: generated/

    texture_of_complex_phase

"""

from .wradlib_bridge import texture_of_complex_phase
from .. import retrieve as _retrieve
_retrieve.texture_of_complex_phase = texture_of_complex_phase



"""
=============================
Utilities (:mod:`pyart.util`)
=============================

Miscellaneous utility functions.

The location and names of these functions within Py-ART may change between
versions without depreciation, use with caution.
.. currentmodule:: pyart.util

Direction statistics
====================
.. autosummary::
    :toctree: generated/
    angular_mean
    angular_std
    angular_mean_deg
    angular_std_deg
    interval_mean
    interval_std
    mean_of_two_angles
    mean_of_two_angles_deg

Miscellaneous functions
=======================
.. autosummary::
    :toctree: generated/

    compute_antenna_diagram
    compute_azimuthal_average
    find_neighbour_gates
    get_target_elevations
    compute_directional_stats
    cross_section_ppi
    cross_section_rhi
    colocated_gates
    colocated_gates2
    intersection
    datetime_from_radar
    datetimes_from_radar
    datetime_from_dataset
    datetimes_from_dataset
    datetime_from_grid
    estimate_noise_hs74
    estimate_noise_ivic13
    get_ivic_pct
    get_ivic_flat_reg_var_max
    get_ivic_snr_thr
    ivic_pct_table
    ivic_snr_thr_table
    ivic_flat_reg_var_max_table
    ivic_flat_reg_wind_len_table
    is_vpt
    to_vpt
    join_radar
    join_spectra
    cut_radar
    cut_radar_spectra
    radar_from_spectra
    interpol_spectra
    ma_broadcast_to
    simulated_vel_from_profile
    texture_along_ray
    texture
    rolling_window
    angular_texture_2d
    compute_nse
    compute_corr
    compute_mse

"""

from .circular_stats import angular_mean, angular_std
from .circular_stats import angular_mean_deg, angular_std_deg
from .circular_stats import interval_mean, interval_std
from .circular_stats import mean_of_two_angles, mean_of_two_angles_deg
from .circular_stats import compute_directional_stats
from .datetime_utils import datetime_from_radar, datetimes_from_radar
from .datetime_utils import datetime_from_dataset, datetimes_from_dataset
from .datetime_utils import datetime_from_grid
from .datetime_utils import EPOCH_UNITS
from .xsect import cross_section_ppi, cross_section_rhi, get_target_elevations
from .xsect import colocated_gates, colocated_gates2, intersection
from .hildebrand_sekhon import estimate_noise_hs74
from .ivic import estimate_noise_ivic13, get_ivic_pct
from .ivic import get_ivic_flat_reg_var_max, get_ivic_snr_thr
from .ivic import ivic_pct_table, ivic_snr_thr_table
from .ivic import ivic_flat_reg_var_max_table, ivic_flat_reg_wind_len_table
from .radar_utils import is_vpt, to_vpt, join_radar, join_spectra
from .radar_utils import cut_radar, cut_radar_spectra, radar_from_spectra
from .radar_utils import interpol_spectra, find_neighbour_gates
from .radar_utils import compute_azimuthal_average, ma_broadcast_to
from .simulated_vel import simulated_vel_from_profile
from .sigmath import texture_along_ray, rolling_window
from .sigmath import texture, angular_texture_2d, grid_texture_2d
from .sigmath import compute_nse, compute_corr, compute_mse

__all__ = [s for s in dir() if not s.startswith('_')]



"""
=====================================================
Dealias doppler velocities using the FourDD algorithm
=====================================================

In this example doppler velocities are dealiased using the Univ. of
Washington FourDD algorithm implemented in Py-ART.  Sounding data is
used for the initial condition of the dealiasing.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import netCDF4
import pyart

SOND_NAME = 'sgpinterpolatedsondeC1.c1.20110510.000000.cdf'
RADAR_NAME = '095636.mdv'

# read in the data
radar = pyart.io.read_mdv(RADAR_NAME)

# read in sonde data
dt, profile = pyart.io.read_arm_sonde_vap(SOND_NAME, radar=radar)

# create a gate filter which specifies gates to exclude from dealiasing
gatefilter = pyart.filters.GateFilter(radar)
gatefilter.exclude_transition()
gatefilter.exclude_invalid('velocity')
gatefilter.exclude_invalid('reflectivity')
gatefilter.exclude_outside('reflectivity', 0, 80)

# perform dealiasing
dealias_data = pyart.correct.dealias_fourdd(
    radar, sonde_profile=profile, gatefilter=gatefilter)
radar.add_field('corrected_velocity', dealias_data)

# create a plot of the first and sixth sweeps
fig = plt.figure(figsize=(15, 10))
ax1 = fig.add_subplot(221)
display = pyart.graph.RadarDisplay(radar)
display.plot('velocity', 0, vmin=-16, vmax=16, ax=ax1,
             colorbar_label='', title='Raw Doppler Velocity, First Sweep')

ax2 = fig.add_subplot(222)
display.plot('corrected_velocity', 0, vmin=-40, vmax=40,
             colorbar_label='', ax=ax2,
             title='Corrected Doppler Velocity, First Sweep')

ax3 = fig.add_subplot(223)
display = pyart.graph.RadarDisplay(radar)
display.plot('velocity', 5, vmin=-16, vmax=16,
             colorbar_label='', ax=ax3,
             title='Raw Doppler Velocity, Sixth Sweep')

ax4 = fig.add_subplot(224)
display.plot_ppi('corrected_velocity', 5, vmin=-40, vmax=40,
                 colorbar_label='', ax=ax4,
                 title='Corrected Doppler Velocity, Sixth Sweep')
plt.suptitle('Velocity dealiasing using Py-ART', fontsize=16)
plt.show()



"""
==================================
Create a PPI plot on a cartopy map
==================================

An example which creates a PPI plot of a file with a cartopy background
and range rings

"""

print(__doc__)

#Author: Jason Hemedinger
#License: BSD 3 clause

import numpy as np
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import pyart

# Read in the file, create a RadarMapDisplay object
filename = 'nsaxsaprppiC1.a1.20140201.184802.nc'
radar = pyart.io.read(filename)
display = pyart.graph.RadarMapDisplay(radar)

# Setting projection and ploting the second tilt
projection = ccrs.LambertConformal(central_latitude=radar.latitude['data'][0],
                                   central_longitude=radar.longitude['data'][0])

fig = plt.figure(figsize=(6,6))
display.plot_ppi_map('reflectivity_horizontal', 1, vmin=-20, vmax=20,
                     min_lon=-157.1, max_lon=-156, min_lat=71.2, max_lat=71.6,
                     lon_lines=np.arange(-158, -154, .2), resolution='10m',
                     lat_lines=np.arange(69, 72, .1), projection=projection,
                     fig=fig, lat_0=radar.latitude['data'][0],
                     lon_0=radar.longitude['data'][0])

# Plot range rings at 10, 20, 30, 40km
display.plot_range_ring(10., line_style='k-')
display.plot_range_ring(20., line_style='k--')
display.plot_range_ring(30., line_style='k-')
display.plot_range_ring(40., line_style='k--')

# Plot cross hairs
display.plot_line_xy(np.array([-40000.0, 40000.0]), np.array([0.0, 0.0]),
                     line_style='k-')
display.plot_line_xy(np.array([0.0, 0.0]), np.array([-20000.0, 200000.0]),
                     line_style='k-')

# Indicate the radar location with a point
display.plot_point(radar.longitude['data'][0], radar.latitude['data'][0])

plt.show()



from __future__ import division, absolute_import, print_function

import sys
from nose import SkipTest

def test_import():
    if sys.version_info[0] >= 3:
        raise SkipTest("traitsdoc not ported to Python3")
    import numpydoc.traitsdoc

# No tests at the moment...
